"""
Module for creating different PowerGAMA scenarios by scaling grid model parameters
according to specified input.
"""

import numpy as np
import pandas as pd

import powergama.constants as const

_EMPTY = np.nan


def saveScenario(base_grid_data, scenario_file, verbose=True):
    """
    Saves the data in the current grid model to a scenario file of the
    format used to create new scenarios

    Parameters
    ----------
        base_grid_data :  GridData
            PowerGAMA GridData object
        scenario_file : string
            name of new scenario (CSV) file


    """

    def printV(*args, **kwargs):
        if verbose:
            print(*args, **kwargs)

    areas_grid = base_grid_data.getAllAreas()
    consumers = base_grid_data.getConsumersPerArea()
    generators = base_grid_data.getGeneratorsPerAreaAndType()
    gentypes_grid = base_grid_data.getAllGeneratorTypes()
    data = {}
    data["demand_annual"] = {}
    data["demand_profile"] = {}
    data["demand_flex_fraction"] = {}
    data["demand_flex_on_off"] = {}
    data["demand_flex_storage"] = {}

    for co in areas_grid:
        print("CO=" + co)

        # Demand
        if co in consumers:
            loads_this_area = [base_grid_data.consumer.demand_avg[i] for i in consumers[co]]
            load_profile = [base_grid_data.consumer.demand_ref[i] for i in consumers[co]]
            demand_sum_MW = float(sum(loads_this_area))
            demandprofiles_set = set(load_profile)
            demand_ref = " ".join(str(x) for x in demandprofiles_set)

            printV("  demand_avg={0:12.2f} <> demand_ref={1:1s}".format(demand_sum_MW, demand_ref))
            # avg demand is in MW, whereas input file is GWh
            data["demand_annual"][co] = demand_sum_MW * const.hoursperyear / const.MWh_per_GWh
            data["demand_profile"][co] = demand_ref

            flex_frac_avg = base_grid_data.consumer.loc[consumers[co], "flex_fraction"].mean()
            flex_onoff_avg = base_grid_data.consumer.loc[consumers[co], "flex_on_off"].mean()
            flex_storage_avg = base_grid_data.consumer.loc[consumers[co], "flex_storage"].mean()
            data["demand_flex_fraction"][co] = flex_frac_avg
            data["demand_flex_on_off"][co] = flex_onoff_avg
            data["demand_flex_storage"][co] = flex_storage_avg

        else:
            printV("  -- no consumers -- ")
            data["demand_annual"][co] = ""
            data["demand_profile"][co] = ""
            data["demand_flex_fraction"][co] = ""
            data["demand_flex_on_off"][co] = ""
            data["demand_flex_storage"][co] = ""

        # Generation inflow, capacity and costs
        for gentype in gentypes_grid:
            if co in generators and gentype in generators[co]:
                inflow_this_area = [base_grid_data.generator.inflow_fac[i] for i in generators[co][gentype]]
                inflow_profile = [base_grid_data.generator.inflow_ref[i] for i in generators[co][gentype]]
                inflowprofiles_set = set(inflow_profile)
                inflow_ref = " ".join(str(x) for x in inflowprofiles_set)
                if len(inflow_this_area) > 0:
                    inflow_avg = float(sum(inflow_this_area)) / len(inflow_this_area)
                else:
                    inflow_avg = None

                storagelevel_this_area = [base_grid_data.generator.storage_ini[i] for i in generators[co][gentype]]
                if len(storagelevel_this_area) > 0:
                    storagelevel_avg = float(sum(storagelevel_this_area)) / len(storagelevel_this_area)
                else:
                    storagelevel_avg = None

                storval_filling_refs = [
                    base_grid_data.generator.storval_filling_ref[i] for i in generators[co][gentype]
                ]
                storval_filling_refs_set = set(storval_filling_refs)
                storval_filling_ref = " ".join(str(x) for x in storval_filling_refs_set)
                storval_time_refs = [base_grid_data.generator.storval_time_ref[i] for i in generators[co][gentype]]
                storval_time_refs_set = set(storval_time_refs)
                storval_time_ref = " ".join(str(x) for x in storval_time_refs_set)

                gencap_this_area = [base_grid_data.generator.pmax[i] for i in generators[co][gentype]]
                gencap_MW = float(sum(gencap_this_area))

                storagecap_this_area = [base_grid_data.generator.storage_cap[i] for i in generators[co][gentype]]
                storagecap_MWh = float(sum(storagecap_this_area))

                gencost_this_area = [base_grid_data.generator.fuelcost[i] for i in generators[co][gentype]]
                gencost_avg = float(sum(gencost_this_area)) / len(gencost_this_area)

                storval_this_area = [
                    base_grid_data.generator.storage_price[i]
                    for i in generators[co][gentype]
                    if base_grid_data.generator.storage_price[i] != 0
                ]
                if not storval_this_area:
                    storval_avg = None
                else:
                    storval_avg = float(sum(storval_this_area)) / len(storval_this_area)

                deadband_this_area = [
                    base_grid_data.generator.pump_deadband[i]
                    for i in generators[co][gentype]
                    if base_grid_data.generator.pump_deadband[i] != 0
                ]
                if not deadband_this_area:
                    pump_deadband = None
                else:
                    pump_deadband = float(sum(deadband_this_area)) / len(deadband_this_area)
                pumpcap_this_area = [base_grid_data.generator.pump_cap[i] for i in generators[co][gentype]]
                pump_capacity = float(sum(pumpcap_this_area))

                pumpeff_this_area = [
                    base_grid_data.generator.pump_efficiency[i]
                    for i in generators[co][gentype]
                    if base_grid_data.generator.pump_efficiency[i] != 0
                ]
                if not pumpeff_this_area:
                    pump_efficiency = None
                else:
                    pump_efficiency = float(sum(pumpeff_this_area)) / len(pumpeff_this_area)

                printV(
                    (
                        "  {0:1s}: cap={1:6.0f}, storage={2:1.0f}"
                        + ", fuelcost_avg={3:6.2f},\n    storval_avg={4:1s}"
                        + ", inflow_fac={5:6.2f}, inflow_ref={6:1s}"
                        + ", \n    storval_fill={7:1s} & _time={8:1s}"
                    ).format(
                        gentype,
                        gencap_MW,
                        storagecap_MWh,
                        gencost_avg,
                        str(storval_avg),
                        inflow_avg,
                        inflow_ref,
                        storval_filling_ref,
                        storval_time_ref,
                    )
                )

                # Create (empty) elements if they haven't already been created.
                if ("gencap_%s" % gentype) not in data:
                    data["gencap_%s" % gentype] = {}
                if ("fuelcost_%s" % gentype) not in data:
                    data["fuelcost_%s" % gentype] = {}
                if ("storage_price_%s" % gentype) not in data:
                    data["storage_price_%s" % gentype] = {}
                if ("inflow_profile_%s" % gentype) not in data:
                    data["inflow_profile_%s" % gentype] = {}
                if ("inflow_factor_%s" % gentype) not in data:
                    data["inflow_factor_%s" % gentype] = {}
                if ("storagecap_%s" % gentype) not in data:
                    data["storagecap_%s" % gentype] = {}
                if ("storage_ini_%s" % gentype) not in data:
                    data["storage_ini_%s" % gentype] = {}
                if ("storval_filling_ref_%s" % gentype) not in data:
                    data["storval_filling_ref_%s" % gentype] = {}
                if ("storval_time_ref_%s" % gentype) not in data:
                    data["storval_time_ref_%s" % gentype] = {}
                if ("pump_deadband_%s" % gentype) not in data:
                    data["pump_deadband_%s" % gentype] = {}
                if ("pump_capacity_%s" % gentype) not in data:
                    data["pump_capacity_%s" % gentype] = {}
                if ("pump_efficiency_%s" % gentype) not in data:
                    data["pump_efficiency_%s" % gentype] = {}

                data["gencap_%s" % gentype][co] = gencap_MW
                data["fuelcost_%s" % gentype][co] = gencost_avg
                data["inflow_profile_%s" % gentype][co] = inflow_ref
                data["inflow_factor_%s" % gentype][co] = inflow_avg
                data["storagecap_%s" % gentype][co] = storagecap_MWh
                data["storage_ini_%s" % gentype][co] = storagelevel_avg
                data["storage_price_%s" % gentype][co] = storval_avg
                data["storval_filling_ref_%s" % gentype][co] = storval_filling_ref
                data["storval_time_ref_%s" % gentype][co] = storval_time_ref
                data["pump_deadband_%s" % gentype][co] = pump_deadband
                data["pump_capacity_%s" % gentype][co] = pump_capacity
                data["pump_efficiency_%s" % gentype][co] = pump_efficiency
            else:
                printV("  {0:1s}:None".format(gentype))

        # end collecting data

    # remove empty lines
    keys_delete = []
    for k in data:
        # isPresent = bool([a for a in data[k].values() if a != None])
        isPresent = any(data[k].values())
        if not isPresent:
            printV("DOES NOT HAVE ", k)
            keys_delete.append(k)
    for k in keys_delete:
        printV("ignoring ", k)
        del data[k]

    # print to file
    df = pd.DataFrame.from_dict(data, orient="index")
    df.to_csv(scenario_file, sep=",")

    return data


def newScenario(base_grid_data, scenario_file, newfile_prefix=None):
    """
    Create new dataset by modifying grid model according to scenario file

    This method replaces generator and consumer data according to
    information given in scenario file. Information that should not
    be replaced should be omitted from or have an empty value in the
    scenario file; the default is to keep existing value.

    Parameters
    ----------
        base_grid_data : GridData
            PowerGAMA grid model object used as basis for modifications
        scenario_file : string
            Name of scenario file (CSV)
        newfiles_prefix : string
            Prefix used when creating new files. New files will be
            the same as old files with this additional prefix

    """

    areas_grid = base_grid_data.getAllAreas()
    gentypes_grid = base_grid_data.getAllGeneratorTypes()
    gentypes_data = []

    df = pd.read_csv(scenario_file, index_col=0)
    areas_data = list(df.columns.values)
    datadict = df.transpose().to_dict()

    areas_nodata = list(set(areas_grid).difference(areas_data))
    areas_notingrid = list(set(areas_data).difference(areas_grid))
    areas_update = list(set(areas_grid).intersection(areas_data))
    print("These areas have no scenario data (will use base values):\n" + str(areas_nodata))
    print("These areas (with data) are not present in the model" + "(will be ignored):\n" + str(areas_notingrid))
    print("Data for these areas will be updated:\n" + str(areas_update))

    consumers = base_grid_data.getConsumersPerArea()
    generators = base_grid_data.getGeneratorsPerAreaAndType()

    # copy existing parameters
    load_new = base_grid_data.consumer.demand_avg.copy()
    loadprofiles_new = base_grid_data.consumer.demand_ref.copy()
    inflow_new = base_grid_data.generator.inflow_fac.copy()
    inflowprofiles_new = base_grid_data.generator.inflow_ref.copy()
    gencap_new = base_grid_data.generator.pmax.copy()
    gencost_new = base_grid_data.generator.fuelcost.copy()
    storagecap_new = base_grid_data.generator.storage_cap.copy()
    storagelevel_new = base_grid_data.generator.storage_ini.copy()
    storval_basevalue_new = base_grid_data.generator.storage_price.copy()
    storval_filling_ref_new = base_grid_data.generator.storval_filling_ref.copy()
    storval_time_ref_new = base_grid_data.generator.storval_time_ref.copy()

    for parameter in datadict:
        row = datadict[parameter]
        # print("parameter={}  row.items={}".format(parameter,row.items()))
        # print(row)

        if parameter == "demand_annual":
            print("Annual demand (GWh)")
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            load_new = _scaleDemand(load_new, row, areas_update, consumers)

        elif parameter == "demand_profile":
            row = {k: (_parseId(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Demand profile references")
            loadprofiles_new = _updateDemandProfileRef(loadprofiles_new, row, areas_update, consumers)

        elif parameter in ["demand_flex_fraction", "demand_flex_on_off", "demand_flex_storage"]:
            print("Demand flexibility values")
            for co in areas_update:
                if row[co] is not _EMPTY:
                    new_val = _parseNum(row[co])
                    if co not in consumers:
                        consumers[co] = []
                        print(f"  WARNING No consumers in {co}")
                        print(row)
                    ind_consumer_per_co = consumers[co]
                    pg_param = parameter[7:]  # remove the the "demand_" part
                    base_grid_data.consumer.loc[ind_consumer_per_co, pg_param] = new_val

        # elif parameter[:14] == "inflow_annual_":
        elif parameter[:14] == "inflow_factor_":
            gentype = parameter[14:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Inflow factor for " + str(gentype))
            inflow_new = _updateInflowFactor(inflow_new, row, areas_update, generators, gentype)

        elif parameter[:15] == "inflow_profile_":
            gentype = parameter[15:]
            row = {k: (_parseId(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Inflow profile references for " + str(gentype))
            inflowprofiles_new = _updateGenProfileRef(inflowprofiles_new, row, areas_update, generators, gentype)

        elif parameter[:7] == "gencap_":
            gentype = parameter[7:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Generation capacities for " + str(gentype))
            if gentype not in gentypes_grid:
                print("OBS: Generation type is not present in grid model:", gentype)
            else:
                gentypes_data.append(gentype)
                gencap_new = _scaleGencap(gencap_new, row, areas_update, generators, gentype)

        elif parameter[:9] == "fuelcost_":
            gentype = parameter[9:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Generation fuel costs for " + str(gentype))
            gencost_new = _updateGenCost(gencost_new, row, areas_update, generators, gentype)

        elif parameter[:14] == "storage_price_":
            gentype = parameter[14:]
            # print("row.items={}".format(row.items()))
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Storage base price for " + str(gentype))
            storval_basevalue_new = _updateGenCost(storval_basevalue_new, row, areas_update, generators, gentype)

        elif parameter[:11] == "storagecap_":
            gentype = parameter[11:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Storage capacities for " + str(gentype))
            if gentype not in gentypes_grid:
                print("OBS: Generation type is not present in grid model:" + str(gentype))
            else:
                gentypes_data.append(gentype)
                storagecap_new = _scaleStoragecap(storagecap_new, row, areas_update, generators, gentype)

        elif parameter[:12] == "storage_ini_":
            gentype = parameter[12:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Initial storage filling level for " + str(gentype))
            if gentype not in gentypes_grid:
                print("OBS: Generation type is not present in grid model:" + str(gentype))
            else:
                gentypes_data.append(gentype)
                storagelevel_new = _updateStorageLevel(storagelevel_new, row, areas_update, generators, gentype)

        elif parameter[:20] == "storval_filling_ref_":
            gentype = parameter[20:]
            row = {k: (_parseId(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Storage value filling level profile reference for " + str(gentype))
            storval_filling_ref_new = _updateGenProfileRef(
                storval_filling_ref_new, row, areas_update, generators, gentype
            )

        elif parameter[:17] == "storval_time_ref_":
            gentype = parameter[17:]
            row = {k: (_parseId(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Storage value time profile reference for " + str(gentype))
            storval_time_ref_new = _updateGenProfileRef(storval_time_ref_new, row, areas_update, generators, gentype)

        elif parameter[:14] == "pump_capacity_":
            gentype = parameter[14:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Pump capacity for " + str(gentype))
            print("** WARNING ** - NOT IMPLEMENTED !!")

        elif parameter[:14] == "pump_deadband_":
            gentype = parameter[14:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Pump deadband for " + str(gentype))
            print("** WARNING ** - NOT IMPLEMENTED !!")

        elif parameter[:16] == "pump_efficiency_":
            gentype = parameter[16:]
            row = {k: (_parseNum(x) if x != "" else None) for k, x in row.items() if k in areas_update}
            print("Pump efficiency for " + str(gentype))
            print("** WARNING ** - NOT IMPLEMENTED !!")

        elif parameter[:6] == "IGNORE":
            print("Ignoring parameter " + parameter)
        else:
            print("WARNING! Unknown parameter: " + parameter)
            raise Exception("Unknown parameter: %s" % parameter)

    gentypes_nodata = list(set(gentypes_grid).difference(gentypes_data))
    print("These generator types have no scenario data " + "(using base values):" + str(gentypes_nodata))

    # Updating variables
    base_grid_data.consumer.demand_avg = load_new[:]
    base_grid_data.consumer.demand_ref = loadprofiles_new[:]
    base_grid_data.generator.inflow_fac = inflow_new[:]
    base_grid_data.generator.inflow_ref = inflowprofiles_new[:]
    base_grid_data.generator.pmax = gencap_new[:]
    base_grid_data.generator.fuelcost = gencost_new[:]
    base_grid_data.generator.storage_cap = storagecap_new[:]
    base_grid_data.generator.storage_ini = storagelevel_new[:]
    base_grid_data.generator.storage_price = storval_basevalue_new[:]
    base_grid_data.generator.storval_filling_ref = storval_filling_ref_new[:]
    base_grid_data.generator.storval_time_ref = storval_time_ref_new[:]

    if newfile_prefix is not None:
        base_grid_data.writeGridDataToFiles(prefix=newfile_prefix)
    return base_grid_data


def _scaleDemand(demand, datarow, areas_update, consumers):
    # Find all loads in this area

    print("  Total average demand before = " + str(sum(demand)))
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            demand_annual_GWh = datarow[co]
            demand_avg_MW = demand_annual_GWh * const.MWh_per_GWh / const.hoursperyear

            if co not in consumers:
                consumers[co] = []

            loads_this_area = [demand[i] for i in consumers[co]]
            demand_avg_MW_before = float(sum(loads_this_area))

            if demand_avg_MW == 0:
                scalefactor = 0
            elif len(loads_this_area) == 0:
                # There are no loads in this area
                print("  WARNING There are no loads to scale in area " + str(co))
            elif demand_avg_MW_before == 0:
                print("  WARNING Zero demand, cannot scale up in area " + str(co))
                scalefactor = 1
            else:
                scalefactor = demand_avg_MW / float(demand_avg_MW_before)
                # print "Scale factor in %s is %g" % (co,scalefactor)

            for i in consumers[co]:
                demand[i] = demand[i] * scalefactor

    print("  Total average demand after = " + str(sum(demand)))
    return demand


def _updateDemandProfileRef(load_profile, datarow, areas_update, consumers):
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            load_profile_new = datarow[co]
            if co not in consumers:
                consumers[co] = []
                print("  WARNING No consumers in {}".format(co))
                print(datarow)
            for i in consumers[co]:
                load_profile[i] = load_profile_new

    return load_profile


def _scaleGencap(gencap, datarow, areas_update, generators, gentype):
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            gencap_MW_new = datarow[co]

            if co not in generators:
                generators[co] = {}
            if gentype not in generators[co]:
                generators[co][gentype] = []
            gencap_this_area = [gencap[i] for i in generators[co][gentype]]
            gencap_MW_before = float(sum(gencap_this_area))

            if gencap_MW_new == 0:
                scalefactor = 0
            elif len(gencap_this_area) == 0:
                print(("  WARNING No generators of type {0} in area {1}." + " Cannot scale.").format(gentype, co))
            elif gencap_MW_before == 0:
                print(
                    ("  WARNING Zero capacity for generator type {0:1s} in " + "area {1:1s}. Cannot scale.").format(
                        gentype, co
                    )
                )
                scalefactor = 1
            else:
                scalefactor = gencap_MW_new / float(gencap_MW_before)
                # print "Scale factor for %s in %s = %g" % (gentype,co,scalefactor)

            for i in generators[co][gentype]:
                gencap[i] = gencap[i] * scalefactor

    return gencap


def _scaleStoragecap(storagecap, datarow, areas_update, generators, gentype):
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            storagecap_MW_new = datarow[co]

            if co not in generators:
                generators[co] = {}
            if gentype not in generators[co]:
                generators[co][gentype] = []
            storagecap_this_area = [storagecap[i] for i in generators[co][gentype]]
            storagecap_MW_before = float(sum(storagecap_this_area))

            if storagecap_MW_new == 0:
                scalefactor = 0
            elif len(storagecap_this_area) == 0:
                print(("  WARNING No generators of type {} in area {}." + " Cannot scale.").format(gentype, co))
            elif storagecap_MW_before == 0:
                print("  WARNING Zero capacity for generator type %s in area %s. Cannot scale." % (gentype, co))
                scalefactor = 1
            else:
                scalefactor = storagecap_MW_new / storagecap_MW_before
                # print "Scale factor for %s in %s = %g" % (gentype,co,scalefactor)

            for i in generators[co][gentype]:
                storagecap[i] = storagecap[i] * scalefactor
    return storagecap


def _updateInflowFactor(inflowfactor, datarow, areas_update, generators, gentype):
    """Update inflow factors per type and area"""

    for co in areas_update:
        if datarow[co] is not _EMPTY:
            if co in generators and gentype in generators[co]:
                for i in generators[co][gentype]:
                    # all generators of this type and area are given
                    # same inflow factor
                    inflowfactor[i] = datarow[co]

    return inflowfactor


def _updateStorageLevel(storagelevel, datarow, areas_update, generators, gentype):
    """Update initial storage level per type and area"""

    for co in areas_update:
        if datarow[co] is not _EMPTY:
            if co in generators and gentype in generators[co]:
                for i in generators[co][gentype]:
                    # all generators of this type and area are given same
                    # inflow factor
                    storagelevel[i] = datarow[co]
    return storagelevel


def _updateGenCost(gencost, datarow, areas_update, generators, gentype):
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            gencost_new = datarow[co]
            if co in generators and gentype in generators[co]:
                # and there are generators of this type
                for i in generators[co][gentype]:
                    gencost[i] = gencost_new

    return gencost


def _NOTNEEDED_updateStoragePrice(storageprice, datarow, areas_update, generators, gentype):
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            storageprice_new = datarow[co]
            if co in generators and gentype in generators[co]:
                # and there are generators of this type
                for i in generators[co][gentype]:
                    storageprice[i] = storageprice_new

    return storageprice


def _updateGenProfileRef(profile_ref, datarow, areas_update, generators, gentype):
    """Update profile per area and generator type"""
    for co in areas_update:
        if datarow[co] is not _EMPTY:
            if co in generators and gentype in generators[co]:
                for i in generators[co][gentype]:
                    profile_ref[i] = datarow[co]

    return profile_ref


def _parseId(value):
    """parse ID string/integer and return a string

    This method is used when reading input data in order to not interpret
    an integer node id o e.g. 100 as "100.0", but always as "100"
    """
    try:
        # if it is an integer number
        s = str(int(value))
    except ValueError:
        # if not then it must have been a string (or empty)
        s = value
    return s


def _parseNum(num):
    """parse number and return a float"""
    return float(num)



"""
This file has a bunch of useful scripts for loading data as well as running fault cases.
"""
import pathlib
import sqlite3 as db

import numpy as np
import pandas as pd

import powergama

from .LpFaultProblem import FaultResults, LpFaultProblem


def set_problem_startpoint(lpproblem, storage_init, storage_flexload_init):
    """Update storage of lpproblem
    Based on  _storeResultsAndUpdateStorage here
    https://github.com/sighellan/powergama/blob/main/src/powergama/LpProblemPyomo.py#L591
    """

    # 1. Update generator storage:
    storagecapacity = lpproblem._grid.generator["storage_cap"]
    lpproblem._storage = np.maximum(0, np.minimum(storagecapacity, storage_init))

    # 2. Update flexible load storage
    # TODO: Assert that values are allowed.
    lpproblem._storage_flexload = storage_flexload_init


class ProblemState:
    def __init__(self, lpproblem=None, storage=None, flex_load=None):
        if lpproblem is not None:
            self.storage = lpproblem._storage
            self.storage_flexload = lpproblem._storage_flexload
        elif (storage is None) and (flex_load is None):
            self.storage = storage
            self.storage_flexload = flex_load
        else:
            raise ValueError("can only specify in one way")

    def __str__(self):
        return "storage:\n%s\n\nstorage_flexload:\n%s" % (self.storage, self.storage_flexload)


def load_table_from_res(db_filename, tablename):
    if tablename == "Res_Nodes":
        query = "SELECT * FROM Res_Nodes "
    elif tablename == "Res_Storage":
        query = "SELECT * FROM Res_Storage "
    elif tablename == "Res_FlexibleLoad":
        query = "SELECT * FROM Res_FlexibleLoad "
    elif tablename == "Res_Generators":
        query = "SELECT * FROM Res_Generators "
    else:
        raise Exception(f"Unknown SQL database table {tablename}")
    con = db.connect(db_filename)
    with con:
        cur = con.cursor()
        cur.execute(query)
        rows = cur.fetchall()
        col_names = [desc[0] for desc in cur.description]
        df = pd.DataFrame(rows, columns=col_names)
    # explicitly close to avoid error when running test in temporary directory
    con.close()
    return df


def load_flexload_states_from_res(db_filename):
    return load_table_from_res(db_filename, "Res_FlexibleLoad")


def load_storage_states_from_res(db_filename):
    return load_table_from_res(db_filename, "Res_Storage")


def load_shedding_table(res_file, max_ts):
    res_nodes = load_table_from_res(res_file, "Res_Nodes")
    load_shedding = np.ones((max_ts, np.max(res_nodes["indx"]) + 1)) * np.nan
    for _, rr in res_nodes.iterrows():
        load_shedding[int(rr.timestep), int(rr.indx)] = rr.loadshed
    return load_shedding


def solve_retry(cur_gridmodel, cur_db_loc, solver, storage_init, flexload_init):
    try:
        lp_cur = powergama.LpProblem(cur_gridmodel, lossmethod=1)
        if storage_init is not None:
            set_problem_startpoint(lp_cur, storage_init, flexload_init)
        res_cur = powergama.Results(cur_gridmodel, cur_db_loc, replace=True)
        lp_cur.solve(res_cur, solver=solver)
    except Exception:
        # TODO: print log saying we've relaxed problem
        print("Trying again without minimum power generation constraints.")
        # TODO: How do we track this really?
        # Presumably if less than Pmin, it really should be no power?
        # Relax Pmin constraints
        cur_gridmodel.generator["pmin"] = np.zeros(len(cur_gridmodel.generator))
        lp_cur = powergama.LpProblem(cur_gridmodel, lossmethod=1)
        if storage_init is not None:
            set_problem_startpoint(lp_cur, storage_init, flexload_init)
        res_cur = powergama.Results(cur_gridmodel, cur_db_loc, replace=True)
        lp_cur.solve(res_cur, solver=solver)
    return lp_cur, res_cur


def run_failure_case_LpFaultProblem(
    failure_dir,
    data_dir,
    base_case_storage,
    base_case_flexload,
    get_gridmodel,
    timesteps,
    solver,
    num_steps=1,
    lossmethod=1,
):
    """Run simulation with fault situation

    failure_dir : pathlib.Path

    """
    res_file = pathlib.Path(failure_dir) / "failure_case_combined.sqlite3"
    max_timestep = np.max(timesteps)
    full_timerange = range(max_timestep + num_steps)
    cur_gridmodel = get_gridmodel(data_dir, full_timerange)

    # TODO: Make a version similar to solve_retry?
    lp_fault = LpFaultProblem(
        cur_gridmodel,
        base_case_storage,
        base_case_flexload,
        lossmethod=lossmethod,
        timesteps=timesteps,
        fault_duration=num_steps,
    )
    res_fault = FaultResults(cur_gridmodel, res_file, replace=True)
    lp_fault.solve(res_fault, solver=solver)
    return lp_fault, res_file



import copy
import os

import numpy as np
import pandas as pd

from .specify_storage import load_table_from_res

"""
Functions to get statistics on load shedding from a simulation.
"""


def get_load_shedding_data(res_file, data, verbose=False, timeminmax=None, res_nodes=None, node_subset=None):
    if res_nodes is None:
        res_nodes = load_table_from_res(res_file, "Res_Nodes")

    if node_subset is not None:
        res_nodes = res_nodes[res_nodes.indx.isin(node_subset)]
        node_list = node_subset
    else:
        node_list = list(range(data.numNodes()))

    N_steps = data.timerange[-1] + 1  # Assumes we start at 0
    summary_stats = []
    for idx in node_list:
        if verbose:
            print(f"\nNode: {idx}")
        res_idx = res_nodes[res_nodes.indx == idx]
        if timeminmax is not None:
            N_steps = timeminmax[1] - timeminmax[0] + 1
            res_idx = res_idx[(res_idx.timestep <= timeminmax[1]) & (res_idx.timestep >= timeminmax[0])]
        if not (len(res_idx) == N_steps):
            raise Exception("Programming error:     Wrong length")
        sum_shed = np.sum(res_idx.loadshed)
        if verbose:
            print(f"Sum shed: {sum_shed}")
        mean_shed = sum_shed / N_steps
        if verbose:
            print(f"Mean shed: {mean_shed}")
        num_hours_shed = np.count_nonzero(res_idx.loadshed)
        if verbose:
            print(f"Hours shed: {num_hours_shed}")
        shed_probability = num_hours_shed / N_steps
        if verbose:
            print(f"Probability of shedding: {shed_probability}")
        summary_stats.append(
            {
                "indx": idx,
                "sum_shed": sum_shed,
                "mean_shed": mean_shed,
                "hours_shed": num_hours_shed,
                "shed_prob": shed_probability,
                "yearly_hours_shed": shed_probability * 8760,
            }
        )
    shed_data = pd.DataFrame(summary_stats)

    timestep_shed = res_nodes.groupby("timestep")["loadshed"].agg(np.sum)
    fraction_hours_with_shedding = np.sum(timestep_shed > 0) / len(timestep_shed)
    return fraction_hours_with_shedding, shed_data


def get_sum_demand(consumers, full_profiles, node_list=None):
    if node_list is not None:
        consumers = consumers[consumers.node.isin(node_list)]
    sum_demand = 0
    for ii, row in consumers.iterrows():
        ii_tot_demand = row["demand_avg"] * full_profiles.get_multiplier_demand()[row["demand_ref"]]
        sum_demand += ii_tot_demand
    return sum_demand


def get_summary_stats(
    fhwue, unserved_data, consumers, full_profiles, node_list=None, sdigits=4, verbose=True, years=30
):
    # fhwue: fraction_hours_with_undelivered_energy
    fhwue_year = fhwue * 24 * 365
    nfhwu = np.mean(unserved_data["shed_prob"])
    nfhwu_year = nfhwu * 24 * 365
    sum_demand = get_sum_demand(consumers=consumers, full_profiles=full_profiles, node_list=node_list)
    undelivered_fraction = np.sum(unserved_data["sum_shed"]) / sum_demand
    undelivered = np.sum(unserved_data["sum_shed"]) / (1000 * years)  # GWh per year
    if verbose:
        print("Fraction hours with underdelivery: %s" % np.round(fhwue, sdigits))
        print("Average hours with underdelivery per year: %s" % np.round(fhwue_year, sdigits))
        print("Average over nodes fraction hours with underdelivery: %s" % np.round(nfhwu, sdigits))
        print("Average over nodes hours with underdelivery per year: %s" % np.round(nfhwu_year, sdigits))
        print(
            "Fraction energy underdelivered: %s" % np.round(undelivered_fraction, sdigits)
        )  # unserved energy / tot energy demand
        print("Total energy undelivered (GWh per year): %s" % np.round(undelivered, sdigits))
    res = [fhwue, fhwue_year, nfhwu, nfhwu_year, undelivered_fraction, undelivered]
    return res


def preprocess_failure_simulation(res_nodes_base, failure_sim_directory):
    # Start with the base case array
    # Whenever there is a failure situation, and the load shedding is above zero,
    #     replace those timesteps in the base case
    csv_file = f"{failure_sim_directory}.csv"
    if os.path.isfile(csv_file):
        print("Reading preprocessed simulation from file")
        failure_case = pd.read_csv(csv_file)
        return failure_case
    else:
        print(f"Preprocessing failure simulation {failure_sim_directory}")

    failure_case = copy.deepcopy(res_nodes_base)

    num_situations = len(os.listdir(failure_sim_directory))

    for ii in range(num_situations):
        res_table = load_table_from_res(
            f"{failure_sim_directory}/failure_{ii}/failure_case_combined.sqlite3", "Res_Nodes"
        )
        if len(res_table[res_table["loadshed"] > 0]) > 0:

            min_ts = res_table["timestep"].min()
            max_ts = res_table["timestep"].max()

            start_idx = failure_case[failure_case["timestep"] == min_ts].index.min()
            end_idx = failure_case[failure_case["timestep"] == max_ts].index.max()

            failure_case[start_idx : end_idx + 1] = res_table.drop(columns="fault_start")
    failure_case.to_csv(csv_file, index=False)
    return failure_case



"""
Module containing PowerGAMA LpProblem class

 Power flow equations:

 Linearised ("DC") power flow equation
 Pinj - Bprime * theta = 0
           Bprime = (N-1)x(N-1) matrix (removed ref.bus row/column)
           theta = phase angles (at N-1 buses)
           Pinj = generation - load at node (cf makeSbus)

 Relationship between angles and power flow
 Pb - (D x A) x theta = 0
           theta_j = phase angle node j (excluding ref. node)
           Pb_k = power flow branch k
           D = diag(-b_k) (negative of susceptance on branch k)
           A = Mx(N-1) node-branch incidence (adjacency) matrix
"""

import warnings

import networkx as nx
import numpy as np
import pandas as pd
import pyomo.environ as pyo
import pyomo.opt
from pyomo.contrib import appsi
from tqdm import tqdm

from . import constants as const


class LpProblem(pyo.ConcreteModel):
    """
    Class containing problem definition as a LP problem, and function calls
    to solve the problem

    """

    def _create_sets_and_parameters(self, grid_data):
        """Create pyomo model sets"""
        self.s_node = pyo.Set(ordered=True, initialize=grid_data.node["id"].tolist())
        self.s_branch_ac = pyo.Set(ordered=True, initialize=grid_data.branch.index.tolist())
        self.s_branch_dc = pyo.Set(ordered=True, initialize=grid_data.dcbranch.index.tolist())
        self.s_gen = pyo.Set(ordered=True, initialize=grid_data.generator.index.tolist())
        self.s_gen_pump = pyo.Set(ordered=True, initialize=grid_data.getIdxGeneratorsWithPumping())
        self.s_load = pyo.Set(ordered=True, initialize=grid_data.consumer.index.tolist())
        self.s_load_flex = pyo.Set(ordered=True, initialize=grid_data.getIdxConsumersWithFlexibleLoad())
        self.s_area = pyo.Set(ordered=True, initialize=grid_data.getAllAreas())

        # Mutable parameters
        # Quantities that change from timestep to the next:
        self.p_gen_pmin = pyo.Param(
            self.s_gen,
            within=pyo.Reals,
            default=0,
            mutable=True,
            initialize=grid_data.generator["pmin"].values,
        )
        self.p_gen_pmax = pyo.Param(
            self.s_gen,
            within=pyo.Reals,
            default=0,
            mutable=True,
            initialize=grid_data.generator["pmax"].values,
        )
        self.p_gen_cost = pyo.Param(
            self.s_gen,
            within=pyo.Reals,
            default=0,
            mutable=True,
            initialize=grid_data.generator["fuelcost"].values,
        )
        self.p_genpump_cost = pyo.Param(self.s_gen, within=pyo.Reals, default=0, mutable=True)
        self.p_demand = pyo.Param(self.s_load, within=pyo.Reals, default=0, mutable=True)
        self.p_loadflex_cost = pyo.Param(
            self.s_load_flex,
            within=pyo.Reals,
            default=0,
            mutable=True,
            # initialize=grid_data.consumer.loc[self.s_load_flex, "flex_basevalue"].values,
        )
        if self._lossmethod == 2:
            self.p_branch_ac_power_loss = pyo.Param(self.s_branch_ac, within=pyo.Reals, default=0, mutable=True)
            self.p_branch_dc_power_loss = pyo.Param(self.s_branch_ac, within=pyo.Reals, default=0, mutable=True)

    def _create_variables(self):
        """Create pyomo model variables"""
        self.varAcBranchFlow = pyo.Var(self.s_branch_ac, within=pyo.Reals)
        self.varDcBranchFlow = pyo.Var(self.s_branch_ac, within=pyo.Reals)
        if self._lossmethod == 1:
            self.varAcBranchFlow12 = pyo.Var(self.s_branch_ac, within=pyo.NonNegativeReals)
            self.varAcBranchFlow21 = pyo.Var(self.s_branch_ac, within=pyo.NonNegativeReals)
            self.varDcBranchFlow12 = pyo.Var(self.s_branch_dc, within=pyo.NonNegativeReals)
            self.varDcBranchFlow21 = pyo.Var(self.s_branch_dc, within=pyo.NonNegativeReals)
            self.varLossAc12 = pyo.Var(self.s_branch_ac, within=pyo.NonNegativeReals)
            self.varLossAc21 = pyo.Var(self.s_branch_ac, within=pyo.NonNegativeReals)
            self.varLossDc12 = pyo.Var(self.s_branch_dc, within=pyo.NonNegativeReals)
            self.varLossDc21 = pyo.Var(self.s_branch_dc, within=pyo.NonNegativeReals)
        self.varGeneration = pyo.Var(self.s_gen, within=pyo.NonNegativeReals)
        self.varPump = pyo.Var(self.s_gen_pump, within=pyo.NonNegativeReals)
        self.varCurtailment = pyo.Var(self.s_gen, within=pyo.NonNegativeReals)
        self.varFlexLoad = pyo.Var(self.s_load_flex, within=pyo.NonNegativeReals)
        self.varLoadShed = pyo.Var(self.s_load, within=pyo.NonNegativeReals)
        self.varVoltageAngle = pyo.Var(self.s_node, within=pyo.Reals, initialize=0.0)

    def _create_constraint_powerflow_limit(self, grid_data):
        """Constraint: Power flow limit"""

        def maxflowAc_rule(model, j):
            cap = grid_data.branch.loc[j, "capacity"]
            if not np.isinf(cap):
                expr = pyo.inequality(-cap, model.varAcBranchFlow[j], cap)
            else:
                expr = pyo.Constraint.Skip
            return expr

        def maxflowDc_rule(model, j):
            cap = grid_data.dcbranch.loc[j, "capacity"]
            if not np.isinf(cap):
                expr = pyo.inequality(-cap, model.varDcBranchFlow[j], cap)
            else:
                expr = pyo.Constraint.Skip
            return expr

        self.cMaxFlowAc = pyo.Constraint(self.s_branch_ac, rule=maxflowAc_rule)
        self.cMaxFlowDc = pyo.Constraint(self.s_branch_dc, rule=maxflowDc_rule)

    def _create_constraint_powerloss(self, grid_data):
        """Constraint: flow = flow12-flow21 & powerloss"""
        if self._lossmethod == 1:

            def flowAc_rule(model, j):
                expr = model.varAcBranchFlow[j] == model.varAcBranchFlow12[j] - model.varAcBranchFlow21[j]
                return expr

            def flowDc_rule(model, j):
                expr = model.varDcBranchFlow[j] == model.varDcBranchFlow12[j] - model.varDcBranchFlow21[j]
                return expr

            self.cFlowAc = pyo.Constraint(self.s_branch_ac, rule=flowAc_rule)
            self.cFlowDc = pyo.Constraint(self.s_branch_dc, rule=flowDc_rule)

        # 1b Losses vs flow
        if self._lossmethod == 1:
            # Upper capacity limit, since capacity may be infinit
            clip_mw = 500
            br = grid_data.branch
            lossAcA = br["resistance"] * br["capacity"].clip(upper=clip_mw) / const.baseMVA
            lossAcB = 0

            br = grid_data.dcbranch
            lossDcA = br["resistance"] * br["capacity"].clip(upper=clip_mw) / const.baseMVA
            lossDcB = 0

            def lossAc_rule12(model, j):
                expr = model.varLossAc12[j] == model.varAcBranchFlow12[j] * lossAcA[j] + lossAcB
                return expr

            def lossAc_rule21(model, j):
                expr = model.varLossAc21[j] == model.varAcBranchFlow21[j] * lossAcA[j] + lossAcB
                return expr

            def lossDc_rule12(model, j):
                expr = model.varLossDc12[j] == model.varDcBranchFlow12[j] * lossDcA[j] + lossDcB
                return expr

            def lossDc_rule21(model, j):
                expr = model.varLossDc21[j] == model.varDcBranchFlow21[j] * lossDcA[j] + lossDcB
                return expr

            self.cLossAc12 = pyo.Constraint(self.s_branch_ac, rule=lossAc_rule12)
            self.cLossAc21 = pyo.Constraint(self.s_branch_ac, rule=lossAc_rule21)
            self.cLossDc12 = pyo.Constraint(self.s_branch_dc, rule=lossDc_rule12)
            self.cLossDc21 = pyo.Constraint(self.s_branch_dc, rule=lossDc_rule21)

    def _create_constraint_generator_output(self):
        """Constraint: Generator output limit"""

        # Generator output constraint is not necessary, as lower and upper
        # bounds are set for each timestep in _update_progress. Should not
        # be specified as constraint with with pmax as limit, since e.g.
        # PV may have higher production than generator rating.

        # HGS: Doing it anyway, cf Espen Bødal and Martin Kristiansen
        # TODO: Check that there are no problems with this.

        def genMaxLimit_rule(model, i):
            return model.varGeneration[i] <= self.p_gen_pmax[i]

        def genMinLimit_rule(model, i):
            if self.p_gen_pmin[i].value > 0:
                return model.varGeneration[i] >= self.p_gen_pmin[i]
            else:
                return pyo.Constraint.Skip

        self.cGenMaxLimit = pyo.Constraint(self.s_gen, rule=genMaxLimit_rule)
        self.cGenMinLimit = pyo.Constraint(self.s_gen, rule=genMinLimit_rule)

    def _create_constraint_generator_pump(self, grid_data):
        """Constraint: Pump output limit"""

        def pump_rule(model, g):
            expr = model.varPump[g] <= grid_data.generator.loc[g, "pump_cap"]
            return expr

        self.cPump = pyo.Constraint(self.s_gen_pump, rule=pump_rule)

    def _create_constraint_load_flex(self, grid_data):
        """Constraint: Flexible load limit"""

        def flexload_rule(model, i):
            flexLoadMax = (
                grid_data.consumer.loc[i, "demand_avg"]
                * grid_data.consumer.loc[i, "flex_fraction"]
                / grid_data.consumer.loc[i, "flex_on_off"]
            )
            expr = model.varFlexLoad[i] <= flexLoadMax
            return expr

        self.cFlexload = pyo.Constraint(self.s_load_flex, rule=flexload_rule)

    def _create_constraint_powerbalance(self, grid_data):
        """ConstraintPower balance (power flow equation)  (Pnode = B theta)"""

        def powerbalance_rule(model, n):
            lhs = 0
            for g in self._generators_at_node[n]:
                # this is a generator connected to node n
                lhs += model.varGeneration[g]
                if g in model.s_gen_pump:
                    lhs -= model.varPump[g]
            for lod in self._loads_at_node[n]:
                lhs -= self.p_demand[lod]
                lhs += model.varLoadShed[lod]
                if lod in model.s_load_flex:
                    lhs -= model.varFlexLoad[lod]
            for b in self._dcbranch_to_node[n]:
                lhs += model.varDcBranchFlow[b]
                if model._lossmethod == 1:
                    lhs -= model.varLossDc12[b]
                elif model._lossmethod == 2:
                    lhs -= model.p_branch_dc_power_loss[b] / 2
            for b in self._dcbranch_from_node[n]:
                lhs += -model.varDcBranchFlow[b]
                if model._lossmethod == 1:
                    lhs -= model.varLossDc21[b]
                elif model._lossmethod == 2:
                    lhs -= model.p_branch_dc_power_loss[b] / 2
            if self._lossmethod == 1:
                # add ac branch losses as load
                for b in self._branch_to_node[n]:
                    lhs += -model.varLossAc12[b]
                for b in self._branch_from_node[n]:
                    lhs += -model.varLossAc21[b]
            elif self._lossmethod == 2:
                for b in self._branch_to_node[n]:
                    # positive sign for flow into node
                    lhs -= model.p_branch_ac_power_loss[b] / 2
                for b in self._branch_from_node[n]:
                    lhs -= model.p_branch_ac_power_loss[b] / 2

            lhs = lhs / const.baseMVA

            # self._powerbalance_rhs = self._get_powerbalance_rhs()
            rhs = self._powerbalance_rhs[n]

            expr = lhs == rhs
            # Skip constraint if it is trivial (otherwise run-time error)
            # TODO: Check if this is safe
            if (type(expr) is bool) and (expr is True):
                expr = pyo.Constraint.Skip
            return expr

        self.cPowerbalance = pyo.Constraint(self.s_node, rule=powerbalance_rule)

    def _create_constraint_powerflow_equation(self, grid_data):
        """Constraint: Power balance (power flow vs voltage angle)"""

        # 1.
        def flowangle_rule(model, b):
            lhs = model.varAcBranchFlow[b]
            lhs = lhs / const.baseMVA
            rhs = 0
            # TODO: This can surely be simplified:
            # node id's are strings, but Bbus and DA matrices need matrix indices (int)
            idx_branch = list(self.s_branch_ac).index(b)
            # idx_branch = grid_data.branch.index.get_loc(b)  # Check if this works
            for i in range(len(self._DA[idx_branch].indices)):
                idx_node2 = self._DA[idx_branch].indices[i]
                DA_element = self._DA[idx_branch].data[i]
                n2 = list(model.s_node)[idx_node2]  # list since pyomo set is 1-based (could probably use +1 instead)
                rhs += DA_element * model.varVoltageAngle[n2] * const.baseAngle
            expr = lhs == rhs
            return expr

        self.cFlowAngle = pyo.Constraint(self.s_branch_ac, rule=flowangle_rule)

        # 2. Reference voltag angle)
        def referenceNode_rule(model, n):
            if n in self.refnodes:
                expr = model.varVoltageAngle[n] == 0
            else:
                expr = pyo.Constraint.Skip
            return expr

        self.cReferenceNode = pyo.Constraint(self.s_node, rule=referenceNode_rule)

    def _create_objective(self, grid_data):
        """Create pyomo model objective function"""

        def cost_rule(model):
            """Operational costs: cost of gen, load shed and curtailment"""

            # Operational costs phase 1 (if stage2DeltaTime>0)
            cost = sum(model.varGeneration[i] * self.p_gen_cost[i] for i in model.s_gen)
            cost -= sum(model.varPump[i] * self.p_genpump_cost[i] for i in model.s_gen_pump)
            cost -= sum(model.varFlexLoad[i] * self.p_loadflex_cost[i] for i in model.s_load_flex)
            cost += sum(model.varLoadShed[i] * const.loadshedcost for i in model.s_load)
            return cost

        self.OBJ = pyo.Objective(rule=cost_rule, sense=pyo.minimize)

    def _get_powerbalance_rhs(self):
        """Get rhs expression in powerbalance constraint"""
        # This is taken out of the constraint creation function to speed up constraint creation
        rhs = dict()
        for n in self.s_node:
            rhs[n] = 0
            # node id's are strings, but Bbus and DA matrices need matrix indices (int)
            idx_node = list(self.s_node).index(n)
            for i in range(len(self._Bbus[idx_node].indices)):
                idx_node2 = self._Bbus[idx_node].indices[i]
                B_element = self._Bbus[idx_node].data[i]
                n2 = list(self.s_node)[idx_node2]  # list since pyomo set is 1-based (could probably use +1 instead)
                rhs[n] -= B_element * self.varVoltageAngle[n2] * const.baseAngle
        return rhs

    def __init__(self, grid, lossmethod=0):
        """LP problem formulation

        Parameters
        ==========
        grid : GridData
            grid data object
        lossmethod : int
            loss method; 0=no losses, 1=linearised losses, 2=added as load
        """

        # 1.
        super().__init__()

        # 2. Compute matrices used in power flow equaions
        print("Computing B and DA matrices...")
        self._Bbus, self._DA = grid.compute_power_flow_matrices()

        print("Initialising LP problem...")

        # Helpers
        self._lossmethod = lossmethod
        self._grid = grid
        self.timeDelta = grid.timeDelta
        self._solver_persistent = False
        self._generators_at_node = grid.generator.groupby("node").groups
        self._loads_at_node = grid.consumer.groupby("node").groups
        self._branch_from_node = grid.branch.groupby("node_from").groups
        self._branch_to_node = grid.branch.groupby("node_to").groups
        self._dcbranch_from_node = grid.dcbranch.groupby("node_from").groups
        self._dcbranch_to_node = grid.dcbranch.groupby("node_to").groups
        for n in grid.node["id"]:
            # fill in so dict is defined for all nodes:
            if n not in self._generators_at_node:
                self._generators_at_node[n] = []
            if n not in self._loads_at_node:
                self._loads_at_node[n] = []
            if n not in self._branch_from_node:
                self._branch_from_node[n] = []
            if n not in self._branch_to_node:
                self._branch_to_node[n] = []
            if n not in self._dcbranch_from_node:
                self._dcbranch_from_node[n] = []
            if n not in self._dcbranch_to_node:
                self._dcbranch_to_node[n] = []

        self._idx_generatorsWithPumping = grid.getIdxGeneratorsWithPumping()
        self._idx_generatorsWithStorage = grid.getIdxGeneratorsWithStorage()
        self._idx_consumersWithFlexLoad = grid.getIdxConsumersWithFlexibleLoad()
        self._idx_branchesWithConstraints = grid.getIdxBranchesWithFlowConstraints()
        # self._fancy_progressbar = False

        # Initial values of marginal costs, storage and storage values
        self._storage = (grid.generator["storage_ini"] * grid.generator["storage_cap"]).fillna(0)
        self._storage_flexload = (
            grid.consumer["flex_storagelevel_init"]
            * grid.consumer["flex_storage"]
            * grid.consumer["flex_fraction"]
            * grid.consumer["demand_avg"]
        ).fillna(0)
        self._energyspilled = grid.generator["storage_cap"].copy(deep=True)
        self._energyspilled[:] = 0

        # Find synchronous areas and specify reference node in each area
        G = nx.Graph()
        G.add_nodes_from(grid.node["id"])
        G.add_edges_from(zip(grid.branch["node_from"], grid.branch["node_to"]))
        G_subs = (G.subgraph(c) for c in nx.connected_components(G))
        self.refnodes = []
        for gr in G_subs:
            refnode = list(gr.nodes)[0]
            self.refnodes.append(refnode)
            print("Found synchronous area (size = {}), using ref node = {}".format(gr.order(), refnode))
        # use first node as voltage angle reference

        # 3. Create pyomo model
        self._create_sets_and_parameters(grid)
        self._create_variables()
        self._create_objective(grid)
        self._powerbalance_rhs = self._get_powerbalance_rhs()
        # 3b. Constraints:
        self._create_constraint_powerflow_limit(grid)
        self._create_constraint_powerloss(grid)
        self._create_constraint_generator_output()
        self._create_constraint_generator_pump(grid)
        self._create_constraint_load_flex(grid)
        self._create_constraint_powerbalance(grid)
        self._create_constraint_powerflow_equation(grid)

    def _get_timesteps_to_solve(self, continue_from_last=False, results=None):
        numTimesteps = len(self._grid.timerange)
        time_steps = range(numTimesteps)
        if continue_from_last:
            timestep_last = results.get_last_timestep_in_results()
            # first step is last plus one
            time_steps = range(timestep_last + 1, numTimesteps)
            print(f"Continue from timestep {timestep_last}.")
        return time_steps

    def _relax_and_retry(self, opt, warmstart, count, solve_args):
        raise NotImplementedError

    # TODO: Update to allow persistent model solving
    # (remove+add constraint instead of mutable parameters)
    def _updateLpProblem(self, timestep):
        """
        Function that updates LP problem for a given timestep, due to changed
        power demand, power inflow and marginal generator costs
        """

        # 1. Generator output limits:
        #    -> power output constraints
        P_storage = self._storage / self.timeDelta
        P_max = self._grid.generator["pmax"]
        P_min = self._grid.generator["pmin"]
        for i in self.s_gen:
            # inflow_factor = self._grid.generator.loc[i, "inflow_fac"]
            # capacity = self._grid.generator.loc[i, "pmax"]
            # inflow_profile = self._grid.generator.loc[i, "inflow_ref"]
            # P_inflow = capacity * inflow_factor * self._grid.profiles.loc[timestep, inflow_profile]
            P_inflow = self._grid.getGeneratorAvailablePower(i, timestep)
            if i not in self._idx_generatorsWithStorage:
                """
                Don't let P_max limit the output (e.g. solar PV)
                This won't affect fuel based generators with zero storage,
                since these should have inflow=p_max in any case
                """
                if P_min[i] > 0:
                    self.p_gen_pmin[i] = min(P_inflow, P_min[i])
                self.p_gen_pmax[i] = P_inflow
            else:
                # generator has storage
                if P_min[i] > 0:
                    self.p_gen_pmin[i] = min(max(0, P_inflow + P_storage[i]), P_min[i])
                self.p_gen_pmax[i] = min(max(0, P_inflow + P_storage[i]), P_max[i])

        # TODO: re-create constraint - if persistent solver

        # 2. Update demand
        #    -> power balance constraint
        for i in self.s_load:
            average = self._grid.consumer.loc[i, "demand_avg"] * (1 - self._grid.consumer.loc[i, "flex_fraction"])
            profile_ref = self._grid.consumer.loc[i, "demand_ref"]
            demand_now = self._grid.profiles.loc[timestep, profile_ref] * average
            self.p_demand[i] = demand_now

        # 3. Cost parameters
        #    -> update objective function

        # 3a. generators with storage (storage value)
        for i in self._idx_generatorsWithStorage:
            this_type_filling = self._grid.generator.loc[i, "storval_filling_ref"]
            this_type_time = self._grid.generator.loc[i, "storval_time_ref"]
            storagecapacity = self._grid.generator.loc[i, "storage_cap"]
            fillinglevel = self._storage[i] / storagecapacity
            filling_col = int(round(fillinglevel * 100))
            storagevalue = (
                self._grid.generator.loc[i, "storage_price"]
                * self._grid.storagevalue_filling.loc[filling_col, this_type_filling]
                * self._grid.storagevalue_time.loc[timestep, this_type_time]
            )
            self.p_gen_cost[i] = storagevalue
            if i in self._idx_generatorsWithPumping:
                deadband = self._grid.generator.pump_deadband[i]
                self.p_genpump_cost[i] = storagevalue - deadband

        # 3b. flexible load (storage value)
        for i in self._idx_consumersWithFlexLoad:
            this_type_filling = self._grid.consumer.loc[i, "flex_storval_filling"]
            this_type_time = self._grid.consumer.loc[i, "flex_storval_time"]
            # Compute storage capacity in Mwh (from value in hours)
            storagecapacity_flexload = (
                self._grid.consumer.loc[i, "flex_storage"]  # h
                * self._grid.consumer.loc[i, "flex_fraction"]
                * self._grid.consumer.loc[i, "demand_avg"]
            )  # MW
            fillinglevel = self._storage_flexload[i] / storagecapacity_flexload
            filling_col = int(round(fillinglevel * 100))
            if fillinglevel > 1:
                storagevalue_flex = -const.flexload_outside_cost
            elif fillinglevel < 0:
                storagevalue_flex = const.flexload_outside_cost
            else:
                storagevalue_flex = (
                    self._grid.consumer.flex_basevalue[i]
                    * self._grid.storagevalue_filling.loc[filling_col, this_type_filling]
                    * self._grid.storagevalue_time.loc[timestep, this_type_time]
                )
            self.p_loadflex_cost[i] = storagevalue_flex

        return

    def _update_persistent_model(self, opt):
        """Update objective function, constraints and bounds in persistent model"""
        # Mutable parameters is not enough
        #
        # TODO: Code for persistent solver
        # https://pyomo.readthedocs.io/en/latest/solvers/persistent_solvers.html
        # to use pwersistent solvers, probably have to set instance at the
        # start, and then modify it in each iteration rather than giving
        # it as an argument to opt.solve:
        #    opt.set_instance(self.concretemodel)
        # and then use opt.solve()
        # To modify e.g. a constraint between solves, remove and add, e.g.:
        #    opt.remove_constraint(m.c)
        #    del m.c
        #    m.c = pe.Constraint(expr=m.y <= m.x)
        #    opt.add_constraint(m.c)
        # Variables can be updated without removing/adding
        #    m.x.setlb(1.0)
        #    opt.update_var(m.x)

        # TODO Check that it is correct - speed up possible?
        # TODO check if deleting and recreating constraint expression is necessary
        # (seems not)

        # 1. p_gen_pmin, p_gen_pmax => gen maxmin
        # 2. p_demand, p_branch_ac_power_loss, p_branch_dc_power_loss => powerbalance
        # 3. p_gen_cost, p_genpump_cost, p_loadflex_cost => OBJ

        # 1.
        for c in self.cGenMaxLimit.values():
            opt.remove_constraint(c)
        for c in self.cGenMinLimit.values():
            opt.remove_constraint(c)
        # del self.cGenMaxLimit
        # del self.cGenMinLimit
        # self._create_constraint_generator_output()
        for c in self.cGenMaxLimit.values():
            opt.add_constraint(c)
        for c in self.cGenMinLimit.values():
            opt.add_constraint(c)

        # 2.
        for c in self.cPowerbalance.values():
            opt.remove_constraint(c)
        # del self.cPowerbalance
        # self._create_constraint_powerbalance(self._grid)
        for c in self.cPowerbalance.values():
            opt.add_constraint(c)

        # 3.
        # no remove_object?
        # del self.OBJ
        # self._create_objective(self._grid)
        opt.set_objective(self.OBJ)

    def _updatePowerLosses(self, aclossmultiplier=1, dclossmultiplier=1):
        """Compute power losses from OPF solution and update parameters"""
        if self._lossmethod == 0:
            pass
        elif self._lossmethod == 1:
            # Use constant loss parameters
            # If loss parameters should change, they need to be declared
            # mutable=True
            pass
        elif self._lossmethod == 2:
            # Losses from previous timestep added as load
            for b in self.s_branch_ac:
                #                # r and x are given in pu; theta
                #                loss_pu = r * ((theta_to-theta_from)*const.baseAngle/x)**2
                #                # convert from p.u. to physical unit
                #                lossMVA = loss_pu*const.baseMVA
                # TODO: simpler (check and replace):
                r = self._grid.branch.loc[b, "resistance"]
                lossMVA = r * self.varAcBranchFlow[b] ** 2 / const.baseMVA
                # A multiplication factor to account for reactive current losses
                # (or more precicely, to get similar results as Giacomo in
                # the SmartNet project)
                lossMVA = lossMVA * aclossmultiplier
                self.p_branch_ac_power_loss[b] = lossMVA
            for b in self.s_branch_dc:
                # TODO: Test this before adding
                r_pu = self._grid.dcbranch.loc[b, "resistance"]
                p_pu = self.varDcBranchFlow[b] / const.baseMVA
                loss_pu = r_pu * p_pu**2
                lossMVA = loss_pu * const.baseMVA * dclossmultiplier
                self.p_branch_dc_power_loss[b] = lossMVA
        else:
            raise Exception("Loss method={} is not implemented".format(self._lossmethod))

    def _get_fault_start(self, timestep):
        # Used by LpFaultProblem
        return None

    def _storeResultsAndUpdateStorage(self, timestep, results):
        """Store timestep results in local arrays, and update storage"""

        # 1. Update generator storage:
        inflow_profile_refs = self._grid.generator["inflow_ref"]
        inflow_factor = self._grid.generator["inflow_fac"]
        capacity = self._grid.generator["pmax"]
        pumpedIn = np.zeros(len(capacity))
        energyIn = np.zeros(len(capacity))
        energyOut = np.zeros(len(capacity))
        for i in self.s_gen:
            genInflow = capacity[i] * inflow_factor[i] * self._grid.profiles[inflow_profile_refs[i]][timestep]
            energyIn[i] = genInflow * self.timeDelta
            energyOut[i] = self.varGeneration[i].value * self.timeDelta

        for i in self._idx_generatorsWithPumping:
            Ppump = self.varPump[i].value
            pumpedIn[i] = Ppump * self._grid.generator["pump_efficiency"][i] * self.timeDelta
        energyStorable = self._storage + energyIn + pumpedIn - energyOut
        storagecapacity = self._grid.generator["storage_cap"]
        # self._storage[i] = min(storagecapacity,energyStorable)
        self._storage = np.vstack((storagecapacity, energyStorable)).min(axis=0)
        self._energyspilled = energyStorable - self._storage

        # 2. Update flexible load storage
        for i in self._idx_consumersWithFlexLoad:
            energyIn_flexload = self.varFlexLoad[i].value * self.timeDelta
            energyOut_flexload = (
                self._grid.consumer["flex_fraction"][i] * self._grid.consumer["demand_avg"][i] * self.timeDelta
            )
            self._storage_flexload[i] += energyIn_flexload - energyOut_flexload

        # 3. Collect variable values from optimisation result
        F = self.OBJ()
        Pgen = [self.varGeneration[i].value for i in self.s_gen]
        Ppump = [self.varPump[i].value for i in self.s_gen_pump]
        Pflexload = [self.varFlexLoad[i].value for i in self.s_load_flex]
        Pb = [self.varAcBranchFlow[i].value for i in self.s_branch_ac]
        Pdc = [self.varDcBranchFlow[i].value for i in self.s_branch_dc]
        theta = [self.varVoltageAngle[i].value * const.baseAngle for i in self.s_node]
        # load shedding is aggregated to nodes (due to old code)
        Ploadshed = pd.Series(index=self._grid.node.id, data=[0] * len(self._grid.node.id), dtype=float)
        for j in self.s_load:
            node = self._grid.consumer["node"][j]
            Ploadshed[node] += self.varLoadShed[j].value

        # 4 Collect dual values
        # 4a. branch capacity sensitivity (whether pos or neg flow)
        senseB = []
        for j in self._idx_branchesWithConstraints:
            # for j in self.concretemodel.BRANCH_AC:
            c = self.cMaxFlowAc[j]
            senseB.append(-abs(self.dual[c] / const.baseMVA))
        senseDcB = []
        for j in self.s_branch_dc:
            c = self.cMaxFlowDc[j]
            senseDcB.append(-abs(self.dual[c] / const.baseMVA))

        # 4b. node demand sensitivity (energy balance)
        # TODO: Without abs(...) the value jumps between pos and neg. Why?
        senseN = []
        for j in self.s_node:
            c = self.cPowerbalance[j]
            senseN.append(abs(self.dual[c] / const.baseMVA))

        # consider spilled energy only for generators with storage<infinity
        # energyspilled = zeros(energyStorable.shape)
        # indx = self._grid.getIdxGeneratorsWithNonzeroInflow()
        # energyspilled[indx] = energyStorable[indx]-self._storage[indx]
        energyspilled = self._energyspilled
        storagelevel = self._storage[self._idx_generatorsWithStorage]
        storageprice = [self.p_gen_cost[i].value for i in self._idx_generatorsWithStorage]
        flexload_storagelevel = self._storage_flexload[self._idx_consumersWithFlexLoad]
        flexload_marginalprice = [self.p_loadflex_cost[i].value for i in self._idx_consumersWithFlexLoad]

        # TODO: Only keep track of inflow spilled for generators with
        # nonzero inflow

        # Extract power losses
        if self._lossmethod == 0:
            acPowerLoss = [0] * len(self.s_branch_ac)
            dcPowerLoss = [0] * len(self.s_branch_dc)
        elif self._lossmethod == 1:
            acPowerLoss = [self.varLossAc12[b].value + self.varLossAc21[b].value for b in self.s_branch_ac]
            dcPowerLoss = [self.varLossDc12[b].value + self.varLossDc21[b].value for b in self.s_branch_dc]
        elif self._lossmethod == 2:
            acPowerLoss = list(self.p_branch_ac_power_loss.extract_values().values())
            dcPowerLoss = list(self.p_branch_dc_power_loss.extract_values().values())
        else:
            raise Exception("Lossmethod must be 0,1 or 2")

        results.addResultsFromTimestep(
            timestep=self._grid.timerange[0] + timestep,
            objective_function=F,
            generator_power=Pgen,
            generator_pumped=Ppump,
            branch_power=Pb,
            dcbranch_power=Pdc,
            node_angle=theta,
            sensitivity_branch_capacity=senseB,
            sensitivity_dcbranch_capacity=senseDcB,
            sensitivity_node_power=senseN,
            storage=storagelevel.tolist(),
            inflow_spilled=energyspilled.tolist(),
            loadshed_power=Ploadshed.tolist(),
            marginalprice=storageprice,
            flexload_power=Pflexload,
            flexload_storage=flexload_storagelevel.tolist(),
            flexload_storagevalue=flexload_marginalprice,
            branch_ac_losses=acPowerLoss,
            branch_dc_losses=dcPowerLoss,
            fault_start=self._get_fault_start(timestep),
        )

        return

    def solve(
        self,
        results,
        solver="cbc",
        solver_path=None,
        warmstart=False,
        savefiles=False,
        aclossmultiplier=1,
        dclossmultiplier=1,
        solve_args=None,
        continue_from_last=False,
    ):
        """
        Solve LP problem for each time step in the time range

        Parameters
        ----------
        results : Results
            PowerGAMA Results object reference
        solver : string (optional)
            name of solver to use ("cbc" or "gurobi"). Gurobi uses python
            interface, whilst CBC uses command line executable
        solver_path :string (optional, only relevant for cbc)
            path for solver executable
        warmstart : Boolean
            Use warmstart option (only some solvers, e.g. gurobi)
        savefiles : Boolean
            Save Pyomo model file and LP problem MPS file for each timestep
            This may be useful for debugging.
        aclossmultiplier : float
            Multiplier factor to scale computed AC losses, used with method 1
        dclossmultiplier : float
            Multiplier factor to scale computed DC losses, used with method 1
        logfile : string
            Name of log file for LP solver. Will keep only last iteration
        solve_args : dict
            Arguments passed on to pyomo.solve(...) in each iteration
        continue_from_last : bool
            Whether to continue from last saved result (useful if simulation
            was interrupted)

        Returns
        -------
        results : Results
            PowerGAMA Results object reference
        """
        if solve_args is None:
            solve_args = {
                "tee": False,  # stream the solver output
                "keepfiles": False,  # print the LP file for examination
                "symbolic_solver_labels": True,  # use human readable names
                "logfile": "lpsolver_log.txt",
            }

        if "_persistent" in solver:
            self._solver_persistent = True

        # Initalise solver, and check it is available
        if solver == "gurobi":
            opt = pyo.SolverFactory("gurobi", solver_io="python")
            print(":) Using direct python interface to solver")
        elif solver == "gurobi_direct":  # think this is the same as gurobi above
            opt = pyo.SolverFactory(solver)
            print(":) Using gurobi_direct")
        elif solver == "gurobi_persistent":
            opt = pyo.SolverFactory("gurobi_persistent")
            print(":) Using persistent (in-memory) python interface to solver")
            print("-- Experimental --")
            symbolic_solver_labels = True
            if "symbolic_solver_labels" in solve_args:
                symbolic_solver_labels = solve_args["symbolic_solver_labels"]
                solve_args.pop("symbolic_solver_labels")
            opt.set_instance(self, symbolic_solver_labels=symbolic_solver_labels)
        elif solver == "appsi_highs":
            # opt = pyo.SolverFactory(solver)
            opt = appsi.solvers.Highs()
            if opt.available():
                print(":) Found solver")
            else:
                print(":( Could not find solver {}. Returning.".format(solver))
                raise Exception("Could not find LP solver {}".format(solver))
        else:
            solver_io = None
            # if solver=="cbc":
            # NL requres CBC with ampl interface built in
            #    solver_io="nl"
            opt = pyo.SolverFactory(solver, executable=solver_path, solver_io=solver_io)
            if opt.available():
                print(":) Found solver here: {}".format(opt.executable()))
            else:
                print(":( Could not find solver {}. Returning.".format(solver))
                raise Exception("Could not find LP solver {}".format(solver))

        # Enable access to dual values
        if not isinstance(opt, appsi.solvers.highs.Highs):
            self.dual = pyo.Suffix(direction=pyo.Suffix.IMPORT)

        timesteps_to_solve = self._get_timesteps_to_solve(continue_from_last=continue_from_last, results=results)
        if continue_from_last:
            # TODO update self._storage with value from database (and similar for flex load)
            # this must be done before updateLpProblem below
            self._storage.loc[self._idx_generatorsWithStorage] = results.db.getResultStorageFillingAll(
                timestep=timesteps_to_solve[0]
            )
            self._storage_flexload.loc[self._idx_consumersWithFlexLoad] = results.db.getResultFlexloadStorageFillingAll(
                timestep=timesteps_to_solve[0]
            )
            raise NotImplementedError("CHECK THIS FIRST - and add test case")

        if self._lossmethod == 2:
            print("Computing losses in first timestep")
            self._updateLpProblem(timestep=timesteps_to_solve[0])
            res = opt.solve(self)
            # Now, power flow values are computed for the first timestep, and
            # power losses can be computed.

        print("Solving...")
        count = 0
        warmstart_now = False
        for timestep in tqdm(timesteps_to_solve):
            # update LP problem (inflow, storage, profiles)
            self._updateLpProblem(timestep)
            self._updatePowerLosses(aclossmultiplier, dclossmultiplier)
            if self._solver_persistent:
                self._update_persistent_model(opt=opt)

            # solve the LP problem
            if savefiles:
                # self.concretemodel.pprint('concretemodel_{}.txt'.format(timestep))
                self.write(
                    "LPproblem_{}.mps".format(timestep),
                    io_options={"symbolic_solver_labels": True},
                )
                # self.concretemodel.write("LPproblem_{}.nl".format(timestep))

            if warmstart:
                if opt.warm_start_capable():
                    # warmstart available (does not work with cbc)
                    if count > 0:
                        warmstart_now = warmstart
                    count = count + 1
                    solve_args["warmstart"] = warmstart_now
                else:
                    raise Exception("Solver ({}) is not capable of warm start".format(opt.name))

            try:
                res = opt.solve(self, **solve_args)
            except Exception as ex:
                # do something ()
                print(f"SOLVE ERROR at timestep={timestep}. Tries to solve again")
                try:
                    res = opt.solve(self, **solve_args)
                except Exception:
                    # re-raise exception:
                    raise ex

            # store result for inspection if necessary
            self.solver_res = res

            # debugging:
            if False:
                print(
                    "Solver status = {}. Termination condition = {}".format(
                        res.solver.status, res.solver.termination_condition
                    )
                )

            if isinstance(res, appsi.solvers.highs.HighsResults):
                if res.termination_condition != appsi.base.TerminationCondition.optimal:
                    print("APPSI HIGHS: Non-optimal solution")
                self.dual = opt.get_duals()
            elif res.solver.status != pyomo.opt.SolverStatus.ok:
                warnings.warn("Something went wrong with LP solver: {}".format(res.solver.status))
                try:
                    self._relax_and_retry(opt, warmstart, count, solve_args)
                except NotImplementedError:
                    raise Exception("Something went wrong with LP solver: {}".format(res.solver.status))
            elif res.solver.termination_condition == pyomo.opt.TerminationCondition.infeasible:
                warnings.warn("t={}: No feasible solution found.".format(timestep))
                try:
                    self._relax_and_retry(opt, warmstart, count, solve_args)
                except NotImplementedError:
                    raise Exception("t={}: No feasible solution found.".format(timestep))

            # This call is required for fault scenario simulation. Does nothing here.
            self._update_progress(timestep, len(timesteps_to_solve))

            # store results and update storage levels
            self._storeResultsAndUpdateStorage(timestep, results)

        return results

    def _update_progress(self, n=None, maxn=None):
        return



"""
Module containing PowerGAMA GridData class and sub-classes

Grid data and time-dependent profiles
"""

import math
import pathlib
import warnings

import networkx as nx
import numpy
import pandas as pd
import scipy.sparse


class GridData(object):
    """
    Class for grid data storage and import
    """

    # Headers and default values in input files:
    # default=None: column _must_ be present in input file
    keys_powergama = {
        "node": {"id": None, "area": None, "zone": "", "lat": None, "lon": None},
        "branch": {"node_from": None, "node_to": None, "reactance": None, "capacity": None, "resistance": 0},
        "dcbranch": {"node_from": None, "node_to": None, "capacity": None, "resistance": 0},
        "generator": {
            "type": None,
            "desc": "",
            "node": None,
            "pmax": None,
            "pmin": None,
            "fuelcost": None,
            "inflow_fac": None,
            "inflow_ref": None,
            "storage_cap": 0,
            "storage_price": 0,
            "storage_ini": 0,
            "storval_filling_ref": "",
            "storval_time_ref": "",
            "pump_cap": 0,
            "pump_efficiency": 0,
            "pump_deadband": 0,
        },
        "consumer": {
            "node": None,
            "demand_avg": None,
            "demand_ref": None,
            "flex_fraction": 0,
            "flex_on_off": 0,
            "flex_basevalue": 0,
            "flex_storage": 0,
            "flex_storval_filling": "",
            "flex_storval_time": "",
            "flex_storagelevel_init": 0.5,
        },
    }
    dtypes = {
        "node": {"id": str, "area": str},
        "branch": {"node_from": str, "node_to": str, "capacity": float},
        "dcbranch": {"node_from": str, "node_to": str, "capacity": float},
        "generator": {"node": str, "type": str, "pmax": float},
        "consumer": {"node": str},
    }

    def __init__(self):
        """
        Create GridData object with data and methods for import and
        processing of PowerGAMA grid data
        """
        self.node = None
        self.branch = None
        self.dcbranch = None
        self.generator = None
        self.consumer = None
        self.profiles = None
        self.storagevalue_filling = None
        self.storagevalue_time = None
        self.timeDelta = None
        self.timerange = None

        self.CSV_SEPARATOR = None  # automatically detect

    def _myround(x, base=1, method="round"):
        """Round to nearest multiple of base"""
        if method == "round":
            return int(base * round(float(x) / base))
        elif method == "floor":
            return int(base * math.floor(float(x) / base))
        elif method == "ceil":
            return int(base * math.ceil(float(x) / base))
        else:
            raise Exception("Rounding error")

    def read_csv_files(self, file_input, dataset):
        dtype = self.dtypes[dataset]
        if isinstance(file_input, str) or isinstance(file_input, pathlib.Path):
            # Single file
            df = pd.read_csv(file_input, dtype=dtype)
        elif isinstance(file_input, list):
            # Multiple files
            df_list = [pd.read_csv(file, dtype=dtype) for file in file_input]
            df = pd.concat(df_list, ignore_index=True)
        elif file_input is None:
            # Empty dataframe with the right columns
            df = pd.DataFrame(columns=self.keys_powergama[dataset].keys())
        else:
            raise ValueError("Input must be a file name (str) or a list of file names (list of str)")
        return df

    def from_dict(self, data_dict, timedelta):
        """Get grid data from a dictionary of pandas dataframes

        dictionary must contain dataframes for keys:
        node, branch, dcbranch, generator, consumer, profiles, storval_time, storval_filling
        """
        self.node = data_dict["node"].astype(self.dtypes["node"])
        self.branch = data_dict["branch"].astype(self.dtypes["branch"])
        self.dcbranch = data_dict["dcbranch"].astype(self.dtypes["dcbranch"])
        self.generator = data_dict["generator"].astype(self.dtypes["generator"])
        self.consumer = data_dict["consumer"].astype(self.dtypes["consumer"])
        self.profiles = data_dict["profiles"]
        self.storagevalue_time = data_dict["storval_time"]
        self.storagevalue_filling = data_dict["storval_filling"]
        self.timedelta = timedelta
        self.timerange = list(self.profiles.index)

    def readGridData(self, nodes, ac_branches, dc_branches, generators, consumers, remove_extra_columns=False):
        """Read grid data from files into data variables

        nodes: filename or list of filenames or None
        ac_branches: filename or list of filenames or None
        dc_branches: filename or list of filenames or None
        generators: filename or list of filenames or None
        consumers: filename or list of filenames or Non
        """

        self.node = self.read_csv_files(nodes, dataset="node")
        self.branch = self.read_csv_files(ac_branches, dataset="branch")
        self.dcbranch = self.read_csv_files(dc_branches, dataset="dcbranch")
        self.generator = self.read_csv_files(generators, dataset="generator")
        self.consumer = self.read_csv_files(consumers, dataset="consumer")

        self._checkGridDataFields(self.keys_powergama)
        self._checkGridData()
        self._addDefaultColumns(keys=self.keys_powergama, remove_extra_columns=remove_extra_columns)
        self._fillEmptyCells(keys=self.keys_powergama)
        self._checkConsistency()

    def _fillEmptyCells(self, keys):
        """Use default data where none is given"""
        # generators:
        for col, val in keys["generator"].items():
            if val is not None:
                self.generator[col] = self.generator[col].fillna(keys["generator"][col])
        # consumers:
        for col, val in keys["consumer"].items():
            if val is not None:
                self.consumer[col] = self.consumer[col].fillna(keys["consumer"][col])

        # branches:
        for col, val in keys["branch"].items():
            if val is not None:
                self.branch[col] = self.branch[col].fillna(keys["branch"][col])

    def _addDefaultColumns(self, keys, remove_extra_columns=False):
        """insert optional columns with default values when none
        are provided in input files"""
        for k in keys["generator"]:
            if k not in self.generator.keys():
                self.generator[k] = keys["generator"][k]
        for k in keys["consumer"]:
            if k not in self.consumer.keys():
                self.consumer[k] = keys["consumer"][k]
        for k in keys["branch"]:
            if k not in self.branch.keys():
                self.branch[k] = keys["branch"][k]
        for k in keys["dcbranch"]:
            if k not in self.dcbranch.keys():
                self.dcbranch[k] = keys["dcbranch"][k]
        for k in keys["node"]:
            if k not in self.node.keys():
                self.node[k] = keys["node"][k]

        # Discard extra columns (comments etc)
        if remove_extra_columns:
            self.node = self.node[list(keys["node"].keys())]
            self.branch = self.branch[list(keys["branch"].keys())]
            self.dcbranch = self.dcbranch[list(keys["dcbranch"].keys())]
            self.generator = self.generator[list(keys["generator"].keys())]
            self.consumer = self.consumer[list(keys["consumer"].keys())]

    def _checkGridDataFields(self, keys):
        """check if all required columns are present
        (ie. all columns with no default value)"""
        for k, v in keys["node"].items():
            if v is None and k not in self.node:
                raise Exception("Node input file must contain %s" % k)
        for k, v in keys["branch"].items():
            if v is None and k not in self.branch:
                raise Exception("Branch input file must contain %s" % k)
        for k, v in keys["dcbranch"].items():
            if v is None and k not in self.dcbranch:
                raise Exception("DC branch input file must contain %s" % k)
        for k, v in keys["generator"].items():
            if v is None and k not in self.generator:
                raise Exception("Generator input file must contain %s" % k)
        for k, v in keys["consumer"].items():
            if v is None and k not in self.consumer:
                raise Exception("Consumer input file must contain %s" % k)

    def _checkConsistency(self):
        """Check consistency between default and provided values.
        Currently only checks for pumping capacity and efficiency,
        but more tests could be added."""

        # generators
        for g_idx in range(len(self.generator)):
            # Check whether pump capacity and efficiency are consistent
            cur_gen = self.generator.iloc[g_idx]
            if not math.isclose(cur_gen.pump_cap, 0, abs_tol=1e-8) and math.isclose(
                cur_gen.pump_efficiency, 0, abs_tol=1e-8
            ):
                warn_message = (
                    "Warning, generator at node '%s' " % cur_gen.node
                    + "has very low pump efficiency (%s), " % cur_gen.pump_efficiency
                    + "but non-zero pump capacity (%s)." % cur_gen.pump_cap
                )
                warnings.warn(warn_message, UserWarning)

    def _checkGridData(self):
        """Check consistency of grid data"""

        # generator nodes
        for g in self.generator["node"]:
            if g not in self.node["id"].values:
                raise Exception("Generator node does not exist: '%s'" % g)

        # consumer nodes
        for c in self.consumer["node"]:
            if c not in self.node["id"].values:
                raise Exception("Consumer node does not exist: '%s'" % c)

        # branch nodes
        for c in self.branch["node_from"]:
            if c not in self.node["id"].values:
                raise Exception("Branch from node does not exist: '%s'" % c)
        for c in self.branch["node_to"]:
            if c not in self.node["id"].values:
                raise Exception("Branch to node does not exist: '%s'" % c)

        # dcbranch nodes
        for c in self.dcbranch["node_from"]:
            if c not in self.node["id"].values:
                raise Exception("DC Branch from node does not exist: '%s'" % c)
        for c in self.dcbranch["node_to"]:
            if c not in self.node["id"].values:
                raise Exception("DC Branch to node does not exist: '%s'" % c)

    def _readProfileFromFile(self, filename, timerange):
        profiles = pd.read_csv(filename, sep=self.CSV_SEPARATOR, engine="python")
        profiles = profiles.loc[timerange]
        profiles.index = range(len(timerange))
        return profiles

    def _readStoragevaluesFromFile(self, filename):
        profiles = pd.read_csv(filename, sep=self.CSV_SEPARATOR, engine="python")
        return profiles

    def readProfileData(self, filename, timerange, storagevalue_filling=None, storagevalue_time=None, timedelta=1.0):
        """Read profile (timeseries) into numpy arrays

        Storage values have both time dependence and filling level dependence
        The dependence on filling level (0-100%), is given as an array with 101 elements
        """

        self.profiles = self._readProfileFromFile(filename, timerange)
        self.timerange = timerange
        self.timeDelta = timedelta
        if storagevalue_filling is not None:
            self.storagevalue_time = self._readProfileFromFile(storagevalue_time, timerange)
            self.storagevalue_filling = self._readStoragevaluesFromFile(storagevalue_filling)
        return

    def writeGridDataToFiles(self, prefix):
        """
        Save data to new input files
        """

        file_nodes = prefix + "node.csv"
        file_branches = prefix + "branch.csv"
        file_consumers = prefix + "consumer.csv"
        file_generators = prefix + "generator.csv"
        file_dcbranch = prefix + "dcbranch.csv"

        sep = self.CSV_SEPARATOR
        if sep is None:
            sep = ","

        self.node.to_csv(file_nodes, sep=sep, index=False)
        self.branch.to_csv(file_branches, sep=sep, index=False)
        self.consumer.to_csv(file_consumers, sep=sep, index=False)
        self.generator.to_csv(file_generators, sep=sep, index=False)
        self.dcbranch.to_csv(file_dcbranch, sep=sep, index=False)
        return

    def numConsumers(self):
        return self.consumer.shape[0]

    def numGenerators(self):
        return self.generator.shape[0]

    def numNodes(self):
        return self.node.shape[0]

    def numBranches(self):
        return self.branch.shape[0]

    def numDcBranches(self):
        if self.dcbranch is not None:
            return self.dcbranch.shape[0]
        else:
            return 0

    def branchFromNodeIdx(self):
        """get node indices for branch FROM node"""
        # return [self.node[self.node['id']==self.branch['node_from'][k]]
        #        .index.tolist()[0] for k in range(self.numBranches())]
        return [self.node[self.node["id"] == b["node_from"]].index.tolist()[0] for i, b in self.branch.iterrows()]

    def branchToNodeIdx(self):
        """get node indices for branch TO node"""
        return [
            self.node[self.node["id"] == self.branch["node_to"][k]].index.tolist()[0]
            for k in self.branch.index.tolist()
        ]

    def dcBranchFromNodeIdx(self):
        """get node indices for dc branch FROM node"""
        return [
            self.node[self.node["id"] == self.dcbranch["node_from"][k]].index.tolist()[0]
            for k in self.dcbranch.index.tolist()
        ]

    def dcBranchToNodeIdx(self):
        """get node indices for dc branch TO node"""
        return [
            self.node[self.node["id"] == self.dcbranch["node_to"][k]].index.tolist()[0]
            for k in self.dcbranch.index.tolist()
        ]

    def getGeneratorsAtNode(self, nodeIdx):
        """Indices of all generators attached to a particular node"""
        # indices = [i for i, x in enumerate(self.generator['node'])
        #            if x == self.node['id'][nodeIdx]]
        indices = self.generator["node"][self.generator.loc[:, "node"] == self.node["id"][nodeIdx]].index.tolist()
        return indices

    def getGeneratorsWithPumpAtNode(self, nodeIdx):
        """Indices of all pumps attached to a particular node"""
        # indices = [i for i, x in enumerate(self.generator['node'])
        #            if x == self.node['id'][nodeIdx]
        #            and self.generator['pump_cap'][i]>0]
        indices = self.generator["node"][
            (self.generator.loc[:, "node"] == self.node["id"][nodeIdx]) & (self.generator.loc[:, "pump_cap"] > 0)
        ].index.tolist()
        return indices

    def getLoadsAtNode(self, nodeIdx):
        """Indices of all loads (consumers) attached to a particular node"""
        indices = self.consumer["node"][self.consumer.loc[:, "node"] == self.node["id"][nodeIdx]].index.tolist()
        return indices

    def getLoadsFlexibleAtNode(self, nodeIdx):
        """Indices of all flexible nodes attached to a particular node"""
        indices = self.consumer["node"][
            (self.consumer.loc[:, "node"] == self.node["id"][nodeIdx])
            & (self.consumer.loc[:, "flex_fraction"] > 0)
            & (self.consumer.loc[:, "demand_avg"] > 0)
        ].index.tolist()
        return indices

    def getIdxConsumersWithFlexibleLoad(self):
        """Indices of all consumers with flexible load"""
        idx = [
            i
            for i, v in enumerate(self.consumer["flex_fraction"])
            if v > 0 and v < numpy.inf and self.consumer["demand_avg"][i] > 0
        ]
        return idx

    def getFlexibleLoadStorageCapacity(self, consumer_indx):
        """flexible load storage capacity in MWh"""
        cap = (
            self.consumer["demand_avg"][consumer_indx]
            * self.consumer["flex_fraction"][consumer_indx]
            * self.consumer["flex_storage"][consumer_indx]
        )
        return cap

    def getDcBranchesAtNode(self, nodeIdx, direction):
        """Indices of all DC branches attached to a particular node"""
        if direction == "from":
            indices = [i for i, x in enumerate(self.dcbranch["node_from"]) if x == self.node["id"][nodeIdx]]
        elif direction == "to":
            indices = [i for i, x in enumerate(self.dcbranch["node_to"]) if x == self.node["id"][nodeIdx]]
        else:
            raise Exception("Unknown direction in GridData.getDcBranchesAtNode")
        return indices

    def getDcBranches(self):
        """
        Returns a list with DC branches in the format
        [index,from area,to area]
        """
        hvdcBranches = []
        for idx in self.dcbranch.index.tolist():
            fromNodeIdx = self.node["id"].index(self.dcbranch["node_from"][idx])
            toNodeIdx = self.node.name.index(self.dcbranch["node_to"][idx])
            areaFrom = self.node["area"][fromNodeIdx]
            areaTo = self.node["area"][toNodeIdx]
            hvdcBranches.append([idx, areaFrom, areaTo])
        return hvdcBranches

    def getIdxNodesWithLoad(self):
        """Indices of nodes that have load (consumer) attached to them"""
        # Get index of node associated with all consumer
        # indices = numpy.asarray(self.consumer.nodeIdx(self.node))
        # Return indices only once (unique values)
        # indices = numpy.unique(indices)
        loadnodes = self.node[self.node["id"].isin(self.consumer["node"])]
        indices = numpy.asarray(loadnodes.index)
        return indices

    def getIdxGeneratorsWithStorage(self):
        """Indices of all generators with nonzero and non-infinite storage"""
        idx = [i for i, v in enumerate(self.generator["storage_cap"]) if v > 0 and v < numpy.inf]
        return idx

    def getIdxGeneratorsWithNonzeroInflow(self):
        """Indices of all generators with nonzero inflow"""
        idx = [i for i, v in enumerate(self.generator["inflow_fac"]) if v > 0]
        return idx

    def getIdxGeneratorsWithPumping(self):
        """Indices of all generators with pumping capacity"""
        idx = [i for i, v in enumerate(self.generator["pump_cap"]) if v > 0 and v < numpy.inf]
        return idx

    def getIdxBranchesWithFlowConstraints(self):
        """Indices of branches with less than infinite branch capacity"""
        idx = [i for i, v in enumerate(self.branch["capacity"]) if v < numpy.inf]
        return idx

    def getIdxDcBranchesWithFlowConstraints(self):
        """Indices of DC branches with less than infinite branch capacity"""
        if self.dcbranch is None:
            idx = []
        else:
            idx = [i for i, v in enumerate(self.dcbranch["capacity"]) if v < numpy.inf]
        return idx

    def getIdxBranchesWithLength(self):
        """Indices of branches with specified length"""
        idx = [i for i, v in enumerate(self.branch["length"]) if not math.isnan(v)]
        return idx

    def _susceptancePu(self, baseOhm=1):
        """If impedance is already given in pu, baseOhm should be 1
        If not, well... baseOhm depends on the voltage level, so need to know
        the nominal voltage at the bus to convert from ohm to pu.
        """
        if (self.branch["reactance"] == 0).any():
            raise Exception("Zero branch reactance is not allowed.")
        return [-1 / self.branch["reactance"][i] * baseOhm for i in self.branch.index.tolist()]

    def compute_power_flow_matrices(self, base_Z=1):
        """
        Compute and return dc power flow matrices B' and DA
        Parameters
        ==========
        base_Z : float (impedance should already be in pu.)
                base value for impedance
        Returns
        =======
        (coeff_B, coeff_DA) : scipy.sparse matrices
        """

        # df_branch = pd.DataFrame.from_dict(branches, orient="index")
        branches = self.branch
        susceptance = 1 / branches["reactance"] * base_Z
        nodes = list(self.node["id"])
        # node_ids = nodes
        edge_ids = []
        edges = []
        # for br_id, branch in branches.items():
        for br_id, branch in branches.iterrows():
            b = susceptance[br_id]
            edges.append((branch["node_from"], branch["node_to"], br_id, {"i": br_id, "b": b}))
            edge_ids.append(br_id)

        # MultiDiGraph to allow parallel lines
        G = nx.MultiDiGraph()
        G.add_nodes_from(nodes)
        G.add_edges_from(edges)
        A_incidence_matrix = -nx.incidence_matrix(G, oriented=True, nodelist=nodes, edgelist=edges).T
        A_incidence_matrix = scipy.sparse.csc_matrix(A_incidence_matrix)

        # Diagonal matrix
        D = scipy.sparse.diags(-susceptance, offsets=0)
        # Element-wise multiplication:
        DA = D * A_incidence_matrix

        # Bf constructed from incidence matrix with branch susceptance
        # used as weight (this is quite fast)
        Bf = -nx.incidence_matrix(G, oriented=True, nodelist=nodes, edgelist=edges, weight="b").T
        # See note above for this conversion
        Bf = scipy.sparse.csc_matrix(Bf)
        # Element-wise multiplication:
        Bbus = A_incidence_matrix.T * Bf
        return Bbus, DA

    def getAllAreas(self):
        """Return list of areas included in the grid model"""
        areas = self.node["area"]
        allareas = []
        for co in areas:
            if co not in allareas:
                allareas.append(co)
        return allareas

    def getAllGeneratorTypes(self, sort="fuelcost"):
        """Return list of generator types included in the grid model"""
        gentypes = self.generator.type
        alltypes = []
        if sort is None:
            gentypes = self.generator.type
            alltypes = []
            for ge in gentypes:
                if ge not in alltypes:
                    alltypes.append(ge)
            return alltypes
        elif sort == "fuelcost":
            generators = self.getGeneratorsPerType()
            avg = {ge_k: numpy.mean(self.generator.fuelcost[ge_v]) for ge_k, ge_v in generators.items()}
            sorted_list = [k for k in sorted(avg, key=avg.get, reverse=False)]
            return sorted_list
        else:
            raise Exception("sort must be None or 'fuelcost'")

    def getConsumerAreas(self):
        """List of areas for each consumer"""
        areas = [self.node.area[self.node["id"] == n].tolist()[0] for n in self.consumer["node"]]
        return areas

    def getGeneratorAreas(self):
        """List of areas for each generator"""
        areas = [self.node.area[self.node["id"] == n].tolist()[0] for n in self.generator.node]
        return areas

    def getConsumersPerArea(self):
        """Returns dictionary with indices of loads within each area"""
        consumers = {}
        consumer_areas = self.getConsumerAreas()
        for idx_load in self.consumer.index.tolist():
            area_name = consumer_areas[idx_load]
            if area_name in consumers:
                consumers[area_name].append(idx_load)
            else:
                consumers[area_name] = [idx_load]
        return consumers

    def getGeneratorsPerAreaAndType(self):
        """Returns dictionary with indices of generators within each area"""
        generators = {}
        generator_areas = self.getGeneratorAreas()
        for idx_gen in self.generator.index.tolist():
            gtype = self.generator.type[idx_gen]
            area_name = generator_areas[idx_gen]
            if area_name in generators:
                if gtype in generators[area_name]:
                    generators[area_name][gtype].append(idx_gen)
                else:
                    generators[area_name][gtype] = [idx_gen]
            else:
                generators[area_name] = {gtype: [idx_gen]}
        return generators

    def getGeneratorsPerType(self):
        """Returns dictionary with indices of generators per type"""
        generators = {}
        for idx_gen in self.generator.index.tolist():
            gtype = self.generator["type"][idx_gen]
            if gtype in generators:
                generators[gtype].append(idx_gen)
            else:
                generators[gtype] = [idx_gen]
        return generators

    def getGeneratorsWithPumpByArea(self):
        """
        Returns dictionary with indices of generators with pumps within
        each area
        """
        generators = {}
        for pumpIdx, gen in self.generator.iterrows():
            cap = gen["pump_cap"]
            if cap > 0 and cap < numpy.inf:
                nodeId = self.generator.loc[pumpIdx, "node"]
                nodeIdx = self.node[self.node["id"] == nodeId].index[0]
                areaName = self.node.loc[nodeIdx, "area"]
                if areaName in generators:
                    generators[areaName].append(pumpIdx)
                else:
                    generators[areaName] = [pumpIdx]
        return generators

    def getGeneratorAvailablePower(self, gen_index, timestep=None):
        """Available power for generator, considering variable inflow

        If timestep is None, then return for entire profile
        """
        inflow_factor = self.generator.loc[gen_index, "inflow_fac"]
        capacity = self.generator.loc[gen_index, "pmax"]
        inflow_profile = self.generator.loc[gen_index, "inflow_ref"]
        if timestep is None:
            P_inflow = capacity * inflow_factor * self.profiles.loc[:, inflow_profile]
        else:
            P_inflow = capacity * inflow_factor * self.profiles.loc[timestep, inflow_profile]
        return P_inflow

    def getBranchAreas(self):
        br_witharea = pd.DataFrame()
        br_witharea["area_from"] = self.branch.merge(
            right=self.node[["id", "area"]], how="left", left_on="node_from", right_on="id"
        )["area"]
        br_witharea["area_to"] = self.branch.merge(
            right=self.node[["id", "area"]], how="left", left_on="node_to", right_on="id"
        )["area"]
        return br_witharea

    def getDcBranchAreas(self):
        br_witharea = self.dcbranch
        br_witharea["area_from"] = self.dcbranch.merge(
            right=self.node[["id", "area"]], how="left", left_on="node_from", right_on="id"
        )["area"]
        br_witharea["area_to"] = self.dcbranch.merge(
            right=self.node[["id", "area"]], how="left", left_on="node_to", right_on="id"
        )["area"]
        return br_witharea

    def getInterAreaBranches(self, area_from=None, area_to=None, acdc="ac"):
        """
        Get indices of branches from and/or to specified area(s)

        area_from = area from. Use None (default) to leave unspecifie
        area_to= area to. Use None (default) to leave unspecified
        acdc = 'ac' (default) for ac branches, 'dc' for dc branches
        """

        if area_from is None and area_to is None:
            raise Exception("Either from area or to area (or both) has" + "to be specified)")

        if acdc == "ac":
            br_witharea = self.getBranchAreas()
        elif acdc == "dc":
            br_witharea = self.getDcBranchAreas()

        if area_to is not None:
            mask_pos = (br_witharea["area_to"] == area_to) & (br_witharea["area_from"] != area_to)
            mask_neg = (br_witharea["area_from"] == area_to) & (br_witharea["area_to"] != area_to)
        if area_from is not None:
            mask_pos2 = (br_witharea["area_from"] == area_from) & (br_witharea["area_to"] != area_from)
            mask_neg2 = (br_witharea["area_to"] == area_from) & (br_witharea["area_from"] != area_from)
            if area_to is not None:
                mask_pos = mask_pos & mask_pos2
                mask_neg = mask_neg & mask_neg2
            else:
                mask_pos = mask_pos2
                mask_neg = mask_neg2
        branches_pos = list(br_witharea.index[mask_pos])
        branches_neg = list(br_witharea.index[mask_neg])
        return dict(branches_pos=branches_pos, branches_neg=branches_neg)

    def branchDistances(self, R=6373.0):
        """computes branch distance from node coordinates, resuls in km

        Uses haversine formula

        Parameters
        ----------
        R : radius of the Earth
        """

        # approximate radius of earth in km
        n_from = self.branchFromNodeIdx()
        n_to = self.branchToNodeIdx()
        distance = []
        # get endpoint coordinates and convert to radians
        lats1 = self.node["lat"][n_from].apply(math.radians)
        lons1 = self.node["lon"][n_from].apply(math.radians)
        lats2 = self.node["lat"][n_to].apply(math.radians)
        lons2 = self.node["lon"][n_to].apply(math.radians)
        lats1.index = self.branch.index
        lons1.index = self.branch.index
        lats2.index = self.branch.index
        lons2.index = self.branch.index

        for b in self.branch.index:
            lat1 = lats1[b]
            lon1 = lons1[b]
            lat2 = lats2[b]
            lon2 = lons2[b]

            dlon = lon2 - lon1
            dlat = lat2 - lat1

            a = math.sin(dlat / 2) ** 2 + math.cos(lat1) * math.cos(lat2) * math.sin(dlon / 2) ** 2
            c = 2 * math.atan2(math.sqrt(a), math.sqrt(1 - a))
            # atan2 better than asin: c = 2 * math.asin(math.sqrt(a))
            distance.append(R * c)
        return distance

    def spreadNodeCoordinates(self, radius=0.01, inplace=False):
        """Spread nodes with identical coordinates in a small circle
        with radius r

        Parameters
        ----------
        radius : float
            radius in degrees for the size of the spread
        inplace : boolean
            if true, update GridData object

        Returns
        -------
        coords : array
            lat,lon pandas array for nodes
        """
        coords = self.node[["lat", "lon"]]
        dupl_coords = pd.DataFrame()
        dupl_coords["cumcount"] = coords.groupby(["lat", "lon"]).cumcount()
        dupl_coords["count"] = coords.groupby(["lat", "lon"])["lon"].transform("count")
        coords_new = coords.copy()
        for i, c in coords.iterrows():
            n_sum = dupl_coords.loc[i, "count"]
            if n_sum > 1:
                # there are more nodes with the same coordinates
                n = dupl_coords.loc[i, "cumcount"]
                theta = 2 * math.pi / n_sum
                coords_new.loc[i, "lat"] += radius * math.cos(n * theta)
                coords_new.loc[i, "lon"] += radius * math.sin(n * theta)
        if inplace:
            self.node[["lat", "lon"]] = coords_new
        return coords_new



"""
Module dealing with database IO
"""

import os
import sqlite3 as db

import pandas as pd


class DatabaseBaseClass(object):
    """
    Stripped-down version of class for storing results from PowerGAMA in sqlite databse
    """

    SQLITE_MAX_VARIABLE_NUMBER = 990

    def __init__(self, filename):
        self.filename = os.path.abspath(filename)
        self.sqlite_version = db.sqlite_version
        self.timestep_str = "timestep INT"
        self.timestep_qs = "?"

    def createTables(self, data):
        """
        Create database for PowerGAMA results
        """

        nodes = pd.DataFrame(data=data.node[["id", "area", "zone"]])
        br_from = data.branchFromNodeIdx()
        br_to = data.branchToNodeIdx()
        branches = pd.DataFrame()
        branches["fromIndx"] = br_from
        branches["toIndx"] = br_to

        if os.path.isfile(self.filename):
            # delete existing file
            print('Replacing existing SQLite file "{}"'.format(self.filename))
            os.remove(self.filename)
            # Must use a new file
            # raise IOError('Cannot append existing file. Choose new file name.')
        con = db.connect(self.filename)
        with con:
            # Write grid_data dataframes to database:
            data.node.to_sql("data_node", con, if_exists="replace", index=True)
            data.branch.to_sql("data_branch", con, if_exists="replace", index=True)
            data.dcbranch.to_sql("data_dcbranch", con, if_exists="replace", index=True)
            data.generator.to_sql("data_generator", con, if_exists="replace", index=True)
            data.consumer.to_sql("data_consumer", con, if_exists="replace", index=True)
            data.profiles.to_sql("data_profiles", con, if_exists="replace", index=True)
            if data.storagevalue_filling is not None:
                data.storagevalue_filling.to_sql("data_storval_filling", con, if_exists="replace", index=True)
                data.storagevalue_time.to_sql("data_storval_time", con, if_exists="replace", index=True)
            # tables with integer indices - keep for backwards compatibility
            nodes.to_sql("Grid_Nodes", con, if_exists="replace", index=True, index_label="indx")
            branches.to_sql("Grid_Branches", con, if_exists="replace", index=True, index_label="indx")

            cur = con.cursor()
            cur.execute(f"CREATE TABLE Res_ObjFunc({self.timestep_str}, value DOUBLE)")
            cur.execute(f"CREATE TABLE Res_Branches({self.timestep_str}, indx INT," + "flow DOUBLE, loss DOUBLE)")
            cur.execute(f"CREATE TABLE Res_BranchesSens({self.timestep_str}, indx INT," + "cap_sensitivity DOUBLE)")
            cur.execute(
                f"CREATE TABLE Res_DcBranches({self.timestep_str}, indx INT,"
                + "flow DOUBLE, cap_sensitivity DOUBLE,loss DOUBLE)"
            )
            cur.execute(
                f"CREATE TABLE Res_Nodes({self.timestep_str}, indx INT,"
                + "angle DOUBLE, nodalprice DOUBLE, loadshed DOUBLE)"
            )
            cur.execute(
                f"CREATE TABLE Res_Generators({self.timestep_str}, indx INT," + "output DOUBLE, inflow_spilled DOUBLE)"
            )
            cur.execute(
                f"CREATE TABLE Res_Storage({self.timestep_str}, indx INT," + "storage DOUBLE, marginalprice DOUBLE)"
            )
            cur.execute(f"CREATE TABLE Res_Pumping({self.timestep_str}, indx INT," + "output DOUBLE)")
            cur.execute(
                f"CREATE TABLE Res_FlexibleLoad({self.timestep_str}, indx INT,"
                + "demand DOUBLE, storage DOUBLE, value DOUBLE)"
            )

        return nodes

    def get_grid_data(self):
        """Extract grid data from database, as dictionary"""
        con = db.connect(self.filename)
        data = {}
        with con:
            for k in ["node", "branch", "dcbranch", "generator", "consumer", "profiles"]:
                data[k] = pd.read_sql(f"SELECT * FROM data_{k}", con, index_col="index")  # nosec B608
                data[k].index.name = None

            # Query to check if storage value tables exist
            df = pd.read_sql("SELECT name FROM sqlite_master WHERE type='table' AND name='data_storval_time'", con)
            if not df.empty:
                data["storval_time"] = pd.read_sql("SELECT * FROM data_storval_time", con, index_col="index")
                data["storval_filling"] = pd.read_sql("SELECT * FROM data_storval_filling", con, index_col="index")
            else:
                data["storval_time"] = None
                data["storval_filling"] = None
        return data

    def getTimerange(self):
        """
        Get the timesteps
        """
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute("SELECT timestep FROM Res_ObjFunc")
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def timestep_tuple(self, timestep, fault_start):
        # Base version only uses the timestep
        return (timestep,)

    def appendResults(
        self,
        timestep,
        objective_function,
        generator_power,
        generator_pumped,
        branch_flow,
        dcbranch_flow,
        node_angle,
        sensitivity_branch_capacity,
        sensitivity_dcbranch_capacity,
        sensitivity_node_power,
        storage,
        inflow_spilled,
        loadshed_power,
        marginalprice,
        flexload_power,
        flexload_storage,
        flexload_storagevalue,
        idx_storagegen,
        idx_branchsens,
        idx_pumpgen,
        idx_flexload,
        branch_ac_losses,
        branch_dc_losses,
        fault_start=None,
    ):
        """
        Store results from a given timestep to the database

        Parameters
        ----------
        timestep (int)
            timestep number
        objective_function (float)
            value of objective function
        generator_power (list of floats)
            power output of generators
        generator_pumped (list of floats)
            pumped power for generators
        branch_power (list of floats)
            power flow on branches
        node_angle (list of floats)
            phase angle (relative to node 0) at nodes
        sensitivity_branch_capacity (list of floats)
            sensitivity to branch capacity
        sensitivity_dcbranch_capacity (list of floats)
            sensitivty to DC branch capacity
        sensitivity_node_power (list of floats)
            sensitivity to node power (nodal price)
        storage
            storage filling level of generators
        inflow_spilled (list of floats)
            spilled power inflow of generators
        loadshed_power (list of floats)
            unmet power demand at nodes
        marginalprice
            price of generators with storage
        flexload_power (list of floats)
            flexible load power consumption
        flexload_storage
            storage filling level of flexible load
        flexload_storagevalue
            storage value in flexible load energy storage
        idx_storagegen
            index in generator list of generators with storage
        idx_branchsens
            index in branch list of branches with limited capacity
        idx_pumpgen
            index in generator list of generators with pumping
        idx_flexload
            index in consumer list of flexible loads
        branch_ac_losses : list
            ac branch losses
        branch_dc_losses : list
            dc branch losses
        fault_start  (int)
            extra identifier for when simulations are run for several timesteps
            from multiple starting timepoints
        """

        timestep_tuple = self.timestep_tuple(timestep, fault_start)

        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(f"INSERT INTO Res_ObjFunc VALUES({self.timestep_qs},?)", timestep_tuple + (objective_function,))
            cur.executemany(
                f"INSERT INTO Res_Nodes VALUES({self.timestep_qs},?,?,?,?)",
                tuple(
                    timestep_tuple + (i, node_angle[i], sensitivity_node_power[i], loadshed_power[i])
                    for i in range(len(sensitivity_node_power))
                ),
            )
            cur.executemany(
                f"INSERT INTO Res_Branches VALUES({self.timestep_qs},?,?,?)",
                tuple(timestep_tuple + (i, branch_flow[i], branch_ac_losses[i]) for i in range(len(branch_flow))),
            )
            cur.executemany(
                f"INSERT INTO Res_BranchesSens VALUES({self.timestep_qs},?,?)",
                tuple(
                    timestep_tuple + (idx_branchsens[i], sensitivity_branch_capacity[i])
                    for i in range(len(sensitivity_branch_capacity))
                ),
            )
            cur.executemany(
                f"INSERT INTO Res_DcBranches VALUES({self.timestep_qs},?,?,?,?)",
                tuple(
                    timestep_tuple + (i, dcbranch_flow[i], sensitivity_dcbranch_capacity[i], branch_dc_losses[i])
                    for i in range(len(dcbranch_flow))
                ),
            )
            cur.executemany(
                f"INSERT INTO Res_Generators VALUES({self.timestep_qs},?,?,?)",
                tuple(timestep_tuple + (i, generator_power[i], inflow_spilled[i]) for i in range(len(generator_power))),
            )
            cur.executemany(
                f"INSERT INTO Res_Storage VALUES({self.timestep_qs},?,?,?)",
                tuple(timestep_tuple + (idx_storagegen[i], storage[i], marginalprice[i]) for i in range(len(storage))),
            )
            cur.executemany(
                f"INSERT INTO Res_Pumping VALUES({self.timestep_qs},?,?)",
                tuple(
                    timestep_tuple
                    + (
                        idx_pumpgen[i],
                        generator_pumped[i],
                    )
                    for i in range(len(generator_pumped))
                ),
            )
            cur.executemany(
                f"INSERT INTO Res_FlexibleLoad VALUES({self.timestep_qs},?,?,?,?)",
                tuple(
                    timestep_tuple + (idx_flexload[i], flexload_power[i], flexload_storage[i], flexload_storagevalue[i])
                    for i in range(len(flexload_power))
                ),
            )


class Database(DatabaseBaseClass):
    """
    Class for storing results from PowerGAMA in sqlite databse
    """

    def getGridBranches(self):
        """Get branch indices as a list"""
        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            cur = con.cursor()
            cur.execute("SELECT indx,fromIndx,toIndx FROM Grid_Branches ")
            rows = cur.fetchall()
            values = {
                "indx": [row[0] for row in rows],
                "fromIndx": [row[1] for row in rows],
                "toIndx": [row[2] for row in rows],
            }
        return values

    def getResultNodalPrice(self, nodeindx, timeMaxMin):
        """Get nodal price at specified node"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT nodalprice FROM Res_Nodes WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], nodeindx),
            )
            rows = cur.fetchall()
        values = [row[0] for row in rows]
        return values

    def getResultNodalPricesAll(self, timeMaxMin):
        """
        Get nodal price at all nodes (list of tuples)

        Returns
        =======
        List of tuples with values:
        (timestep, node index, nodal price)

        """
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT timestep,indx,nodalprice FROM Res_Nodes"
                " WHERE timestep>=? AND timestep<?"
                " ORDER BY indx,timestep",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
        return rows

    def getResultNodalPricesMean(self, timeMaxMin):
        """Get average nodal price at all nodes"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT indx,AVG(nodalprice) FROM Res_Nodes"
                " WHERE timestep>=? AND timestep<?"
                " GROUP BY indx ORDER BY indx",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
        values = [row[1] for row in rows]
        return values

    def getResultAreaPrices(self, node_weight, timeMaxMin):
        """Get area price timeseries

        node_weight = list of weights for each node
        """

        # print("weight sum = "+str(sum(node_weight)))
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()

            # Temporary table to store weights
            cur.execute("CREATE TABLE IF NOT EXISTS weights" + " (node INT, weight DOUBLE)")
            val = tuple((i, w) for i, w in enumerate(node_weight))
            cur.executemany("INSERT INTO weights VALUES(?,?)", val)

            cur.execute(
                "SELECT n.timestep,SUM(n.nodalprice * w.weight)"
                " FROM Res_Nodes n INNER JOIN weights w ON n.indx==w.node"
                " WHERE n.timestep>=? AND n.timestep<?"
                " GROUP BY n.timestep ORDER BY n.timestep",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
            cur.execute("DROP TABLE weights")
        values = [row[1] for row in rows]
        return values

    def getResultBranchFlow(self, branchindx, timeMaxMin, ac=True):
        """Get branch flow at specified branch"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            if ac:
                table = "Res_Branches"
            else:
                table = "Res_DcBranches"
            cur.execute(
                f"SELECT flow FROM {table} WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1], branchindx),
            )
            rows = cur.fetchall()
        values = [row[0] for row in rows]
        return values

    def getResultBranchFlowAll(self, timeMaxMin, acdc="ac"):
        """
        Get branch flow at all branches (list of tuples)

        Returns
        =======
        List of tuples with values:
        (timestep, branch index, flow)

        """
        valid_tables = {"ac": "Res_Branches", "dc": "Res_DcBranches"}
        if acdc not in valid_tables:
            raise Exception('branch type must be "ac" or "dc"')
        table = valid_tables[acdc]

        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                f"SELECT timestep,indx,flow FROM {table} WHERE timestep>=? AND timestep<? ORDER BY indx,timestep",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
        return rows

    def getResultBranchFlowsMean(self, timeMaxMin, ac=True):
        """
        Get average branch flow on branches in both direction

        Parameters
        ----------
        timeMaxMin (list of two elements) - time interval
        ac (bool) - ac (true) or dc (false) branches

        Returns
        =======
        List with values for each branch:
        [average flow 1->2, average flow 2->1, average absolute flow]

        """
        if ac:
            table = "Res_Branches"
        else:
            table = "Res_DcBranches"

        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                f"""SELECT indx,TOTAL(flow) FROM {table}
                WHERE timestep>=? AND timestep<? AND flow>=0
                GROUP BY indx ORDER BY indx""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows1 = cur.fetchall()
            cur.execute(
                f"""SELECT indx,TOTAL(flow) FROM {table}
                WHERE timestep>=? AND timestep<? AND flow<0
                GROUP BY indx ORDER BY indx""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows2 = cur.fetchall()
            cur.execute(f"SELECT MAX(indx) FROM {table}")  # nosec B608 - is safe even if bandit says no
            numBranches = 1 + cur.fetchone()[0]
            # Calculate average flow for each direction
            numTimeSteps = timeMaxMin[-1] - timeMaxMin[0]
            rows1 = [(index, tot_flow / numTimeSteps) for (index, tot_flow) in rows1]
            rows2 = [(index, tot_flow / numTimeSteps) for (index, tot_flow) in rows2]
        # The length of rows1 and rows2 may be less than the number
        # of branches if the flow is always in one direction
        values_pos = [0] * numBranches
        values_neg = [0] * numBranches
        values_abs = [0] * numBranches
        i1 = 0
        i2 = 0
        for i in range(numBranches):
            if i1 < len(rows1) and rows1[i1][0] == i:
                values_pos[i] = rows1[i1][1]
                i1 = i1 + 1
            if i2 < len(rows2) and rows2[i2][0] == i:
                values_neg[i] = abs(rows2[i2][1])
                i2 = i2 + 1
            values_abs[i] = values_pos[i] + values_neg[i]
        values = [values_pos, values_neg, values_abs]
        return values

    def getResultBranchSens(self, branchindx, timeMaxMin, acdc="ac"):
        """Get branch capacity sensitivity at specified branch"""
        valid_tables = {"ac": "Res_BranchesSens", "dc": "Res_DcBranches"}
        if acdc not in valid_tables:
            raise Exception('branch type must be "ac" or "dc"')
        branch_table = valid_tables[acdc]
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                f"SELECT cap_sensitivity FROM {branch_table} WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1], branchindx),
            )
            rows = cur.fetchall()
        values = [row[0] for row in rows]
        return values

    def getResultBranchSensAll(self, timeMaxMin):
        """Get branch capacity sensitivity at all branches"""
        branches = self.getGridBranches()
        branchlist = branches["indx"]
        sens = []
        for branch in branchlist:
            this_sens = self.getResultBranchSens(branch, timeMaxMin)
            sens.append(this_sens if this_sens != [] else [None] * (timeMaxMin[1] - timeMaxMin[0]))
        return sens

    def getResultBranchSensMean(self, timeMaxMin, acdc="ac"):
        """Get average sensitivity of all  branches
        acdc = 'ac' or 'dc'
        """
        valid_tables = {"ac": "Res_BranchesSens", "dc": "Res_DcBranches"}
        if acdc not in valid_tables:
            raise Exception('branch type must be "ac" or "dc"')
        branch_table = valid_tables[acdc]
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                f"""SELECT indx,AVG(cap_sensitivity)
                FROM {branch_table}
                WHERE timestep>=? AND timestep<?
                GROUP BY indx ORDER BY indx""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
        values = [row[1] for row in rows]
        return values

    def getAverageInterareaBranchFlow(self, timeMaxMin):
        """
        Get average negative flow, positive flow and total flow of branches
        between different areas

        Returns
        =======
        List of tuples for inter-area branches with following values:
        (indices, fromArea, toArea, average negative flow, average positive
        flow, average flow)
        """

        con = db.connect(self.filename)
        db.enable_callback_tracebacks(True)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT b.indx, fromNode.area, toNode.area"
                " FROM Grid_Branches b"
                " INNER JOIN Grid_Nodes fromNode ON b.fromIndx = fromNode.indx"
                " INNER JOIN Grid_Nodes toNode ON b.toIndx = toNode.indx"
                " WHERE fromNode.area != toNode.area"
            )
            branches = cur.fetchall()

            # fetch flows
            cur.execute(
                "SELECT res.indx, TOTAL(res.flow)"
                " FROM Res_Branches res, Grid_Branches b"
                " INNER JOIN Grid_Nodes fromNode ON b.fromIndx = fromNode.indx"
                " INNER JOIN Grid_Nodes toNode ON b.toIndx = toNode.indx"
                " WHERE fromNode.area != toNode.area"
                " AND res.indx = b.indx AND timestep>=?"
                " AND timestep<? AND res.flow<=0"
                " GROUP BY res.indx",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            flow_negative = cur.fetchall()

            cur.execute(
                "SELECT res.indx, TOTAL(res.flow)"
                " FROM Res_Branches res, Grid_Branches b"
                " INNER JOIN Grid_Nodes fromNode ON b.fromIndx = fromNode.indx"
                " INNER JOIN Grid_Nodes toNode ON b.toIndx = toNode.indx"
                " WHERE fromNode.area != toNode.area"
                " AND res.indx = b.indx AND timestep>=? AND timestep<?"
                " AND res.flow>=0"
                " GROUP BY res.indx",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            flow_positive = cur.fetchall()

            cur.execute(
                "SELECT res.indx, TOTAL(res.flow)"
                " FROM Res_Branches res, Grid_Branches b"
                " INNER JOIN Grid_Nodes fromNode ON b.fromIndx = fromNode.indx"
                " INNER JOIN Grid_Nodes toNode ON b.toIndx = toNode.indx"
                " WHERE fromNode.area != toNode.area"
                " AND res.indx = b.indx AND timestep>=? AND timestep<?"
                " GROUP BY res.indx",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            flow_total = cur.fetchall()

            # calculate average flow

            numTimeSteps = timeMaxMin[-1] - timeMaxMin[0]
            flow_negative = [(index, flow / numTimeSteps) for (index, flow) in flow_negative]
            flow_positive = [(index, flow / numTimeSteps) for (index, flow) in flow_positive]
            flow_total = [(index, flow / numTimeSteps) for (index, flow) in flow_total]

            # Sort results
            # The length of flow lists may be less than the number
            # of branches if the flow is always in one direction
            values = []

            # for all inter-area branches
            for index, indice in enumerate([x for (x, y, z) in branches]):
                # find negative flow
                try:
                    temp_ind = [y[0] for y in flow_negative].index(indice)
                    neg = (flow_negative[temp_ind][1],)
                except ValueError:
                    neg = (0,)
                # find positive flow
                try:
                    temp_ind = [y[0] for y in flow_positive].index(indice)
                    pos = (flow_positive[temp_ind][1],)
                except ValueError:
                    pos = (0,)
                # find total flow
                tot = (flow_total[index][1],)

                values.append(branches[index] + neg + pos + tot)
        return values

    def getBranchesSumFlow(self, branches_pos, branches_neg, timeMaxMin, acdc):
        """
        Return time series for aggregated flow along specified branches

        branches_pos = indices of branches with positive flow direction
        branches_neg = indices of branches with negative flow direction
        timeMaxMin = [start, end]
        acdc = 'ac' or 'dc'

        Note: This function can be used to get net import, but not separate
        import/export values (requires summing positive and negative values
        separately)

        """
        values_pos = []
        values_neg = []
        valid_tables = {"ac": "Res_Branches", "dc": "Res_DcBranches"}
        if acdc not in valid_tables:
            raise Exception('branch type must be "ac" or "dc"')
        branch_table = valid_tables[acdc]

        con = db.connect(self.filename)
        with con:
            if branches_pos:
                cur = con.cursor()
                str_bind = ",".join("?" for _ in branches_pos)
                cur.execute(
                    f"""
                    SELECT SUM(flow) FROM {branch_table}
                    WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                    GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                    (timeMaxMin[0], timeMaxMin[-1]) + tuple(branches_pos),
                )
                rows = cur.fetchall()
                values_pos = [row[0] for row in rows]
            if branches_neg:
                cur = con.cursor()
                str_bind = ",".join("?" for _ in branches_neg)
                cur.execute(
                    f"""
                    SELECT SUM(flow) FROM {branch_table}
                    WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                    GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                    (timeMaxMin[0], timeMaxMin[-1]) + tuple(branches_neg),
                )
                rows = cur.fetchall()
                values_neg = [row[0] for row in rows]
            values = dict(pos=values_pos, neg=values_neg)

        return values

    def getResultPumpPower(self, genindx, timeMaxMin):
        """Get pumping for generators with pumping"""
        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            cur = con.cursor()
            cur.execute(
                "SELECT output FROM Res_Pumping WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], genindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultPumpPowerMultiple(self, genindx, timeMaxMin, negative=True):
        """Get pumping for generators with pumping"""

        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            cur = con.cursor()
            str_bind = ",".join("?" for _ in genindx)
            cur.execute(
                f"""SELECT SUM(output) FROM Res_Pumping
                WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]) + tuple(genindx),
            )
            rows = cur.fetchall()
            if negative:
                values = [-row[0] for row in rows]
            else:
                values = [row[0] for row in rows]
        return values

    def getResultStorageFilling(self, genindx, timeMaxMin):
        """Get storage filling level for storage generators"""
        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            cur = con.cursor()
            cur.execute(
                "SELECT storage FROM Res_Storage WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], genindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultStorageFillingAll(self, timestep):
        """Get storage filling level for all storage generators"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute("SELECT indx,storage FROM Res_Storage WHERE timestep=?  ORDER BY indx", (timestep,))
            rows = cur.fetchall()
            values = [row[1] for row in rows]
        return values

    def getResultStorageFillingMultiple(self, genindx, timeMaxMin, capacity=None):
        """Get storage filling level for multiple storage generators"""
        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            str_bind = ",".join("?" for _ in genindx)
            cur = con.cursor()
            cur.execute(
                f"""SELECT SUM(storage) FROM Res_Storage
                WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]) + tuple(genindx),
            )
            rows = cur.fetchall()
            if capacity:
                values = [row[0] / capacity for row in rows]
            else:
                values = [row[0] for row in rows]
        return values

    def getResultStorageValue(self, storageindx, timeMaxMin):
        """Get storage value for storage generators"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT marginalprice FROM Res_Storage  WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], storageindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultStorageValueMultiple(self, storageindx, timeMaxMin):
        """Get average storage value (marginal price) for multiple
        storage generators"""

        con = db.connect(self.filename)
        with con:
            str_bind = ",".join("?" for _ in storageindx)
            cur = con.cursor()
            cur.execute(
                f"""SELECT AVG(marginalprice) FROM Res_Storage
                WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]) + tuple(storageindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultGeneratorSpilledSums(self, timeMaxMin):
        """Get sum of spilled power for all generator"""

        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT indx,SUM(inflow_spilled) "
                " FROM Res_Generators "
                " WHERE timestep>=? AND timestep<?"
                " GROUP BY indx ORDER BY indx",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
            output = [row[1] for row in rows]
        return output

    def getResultGeneratorSpilled(self, generatorindx, timeMaxMin):
        """Get spilled power time series for specified generator"""

        if not isinstance(generatorindx, list):
            generatorindx = [generatorindx]
        if len(generatorindx) == 0:
            return None
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            str_bind = ",".join("?" for _ in generatorindx)
            cur.execute(
                f"""SELECT timestep,SUM(inflow_spilled)
                FROM Res_Generators
                WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                (timeMaxMin[0], timeMaxMin[-1]) + tuple(generatorindx),
            )
            rows = cur.fetchall()
            output = [row[1] for row in rows]
        return output

    def getResultGeneratorPower(self, generatorindx, timeMaxMin):
        """Get power output time series for specified generator"""

        if not isinstance(generatorindx, list):
            generatorindx = [generatorindx]
        if len(generatorindx) == 0:
            return None

        # If there are many generators we need to worry about the max variable number in sqlite
        maxnum = self.SQLITE_MAX_VARIABLE_NUMBER
        indx_chunks = [generatorindx[i : i + maxnum] for i in range(0, len(generatorindx), maxnum)]
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            # get results for chunks of generators:
            output_chunk = []
            for indx in indx_chunks:
                str_bind = ",".join("?" for _ in indx)
                cur.execute(
                    f"""SELECT timestep,SUM(output) FROM Res_Generators
                    WHERE timestep>=? AND timestep<? AND indx IN ({str_bind})
                    GROUP BY timestep ORDER BY timestep""",  # nosec B608 - is safe even if bandit says no
                    (timeMaxMin[0], timeMaxMin[-1]) + tuple(indx),
                )
                rows = cur.fetchall()
                output_chunk.append([row[1] for row in rows])
            # sum chunks for each timestep:
            output = [sum(i) for i in zip(*output_chunk)]
        return output

    def getResultPumpingSum(self, timeMaxMin, variable="output"):
        """Sum of pumping  per generator"""
        con = db.connect(self.filename)
        with con:
            # cur = con.cursor()
            query = (
                "SELECT indx,SUM(?) FROM Res_Pumping  WHERE timestep>=? AND timestep<? GROUP BY indx ORDER BY indx",
                (variable, timeMaxMin[0], timeMaxMin[-1]),
            )
            df = pd.read_sql_query(query, con)
        return df

    def getResultGeneratorPowerSum(self, timeMaxMin):
        """Sum of generator power output per generator"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT indx,SUM(output) FROM Res_Generators "
                " WHERE timestep>=? AND timestep<?"
                " GROUP BY indx"
                " ORDER BY indx",
                (timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
            values = [row[1] for row in rows]
        return values

    def getResultFlexloadPower(self, consumerindx, timeMaxMin):
        """Get flexible load for consumer with flexible load"""
        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            cur = con.cursor()
            cur.execute(
                "SELECT demand FROM Res_FlexibleLoad  WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], consumerindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultFlexloadStorageFilling(self, consumerindx, timeMaxMin):
        """Get storage filling level for flexible loads"""
        con = db.connect(self.filename)
        with con:
            # con.row_factory = db.Row
            cur = con.cursor()
            cur.execute(
                "SELECT storage FROM Res_FlexibleLoad  WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], consumerindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultFlexloadStorageFillingAll(self, timestep):
        """Get storage filling level for all flexible loads"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute("SELECT indx,storage FROM Res_FlexibleLoad WHERE timestep=?  ORDER BY indx", (timestep,))
            rows = cur.fetchall()
            values = [row[1] for row in rows]
        return values

    def getResultFlexloadStorageValue(self, consumerindx, timeMaxMin):
        """Get storage value for flexible loads"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT value FROM Res_FlexibleLoad  WHERE timestep>=? AND timestep<? AND indx=? ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], consumerindx),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultLoadheddingInArea(self, area, timeMaxMin):
        """Aggregated loadshedding timeseries for specified area"""
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT SUM(loadshed) FROM Res_Nodes "
                " WHERE timestep>=? AND timestep<? AND indx IN "
                " (SELECT indx FROM Grid_Nodes WHERE area IN (?))"
                " GROUP BY timestep"
                " ORDER BY timestep",
                (timeMaxMin[0], timeMaxMin[-1], area),
            )
            rows = cur.fetchall()
            values = [row[0] for row in rows]
        return values

    def getResultLoadheddingSum(self, timeMaxMin, average=False):
        """Sum of loadshedding timeseries per node"""
        if average:
            query = (
                "SELECT indx,AVG(loadshed) FROM Res_Nodes  WHERE timestep>=? AND timestep<? GROUP BY indx ORDER BY indx"
            )
        else:
            query = (
                "SELECT indx,SUM(loadshed) FROM Res_Nodes  WHERE timestep>=? AND timestep<? GROUP BY indx ORDER BY indx"
            )
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(query, (timeMaxMin[0], timeMaxMin[-1]))
            rows = cur.fetchall()
            values = [row[1] for row in rows]
        return values

    def getResultBranchLossesSum(self, timeMaxMin, acdc="ac"):
        """Sum of losses for each time-step time step"""
        sqlTable = "Res_Branches"
        if acdc == "dc":
            sqlTable = "Res_DcBranches"
        con = db.connect(self.filename)
        with con:
            cur = con.cursor()
            cur.execute(
                "SELECT indx,SUM(loss) FROM ? WHERE timestep>=? AND timestep<? GROUP BY timestep",
                (sqlTable, timeMaxMin[0], timeMaxMin[-1]),
            )
            rows = cur.fetchall()
            values = [row[1] for row in rows]
        return values

    def getResultBranches(self, timeMaxMin, br_indx=None, acdc="ac"):
        """Branch results for each time-step

        Parameters
        ==========
        timeMaxMin : [start,end]
            tuple with time window start <= t < end
        br_indx : list (optional)
            list of branches to consider, None=include all
        """
        valid_tables = {"ac": "Res_Branches", "dc": "Res_DcBranches"}
        if acdc not in valid_tables:
            raise Exception('branch type must be "ac" or "dc"')
        table = valid_tables[acdc]
        con = db.connect(self.filename)
        with con:
            if br_indx is None:
                query = f"SELECT * FROM {table} WHERE timestep>=? AND timestep<? GROUP BY indx,timestep"  # nosec B608 - is safe even if bandit says no
                df = pd.read_sql_query(query, con, params=(timeMaxMin[0], timeMaxMin[-1]))
            elif len(br_indx) == 0:
                # Empty dataframe
                df = pd.DataFrame()
            else:
                str_bind = ",".join("?" for _ in br_indx)
                query = f"""SELECT * FROM {table}
                    WHERE timestep>=? AND timestep<?  AND indx IN ({str_bind})
                    GROUP BY indx,timestep"""  # nosec B608 - is safe even if bandit says no
                df = pd.read_sql_query(query, con, params=(timeMaxMin[0], timeMaxMin[-1]) + tuple(br_indx))
        return df



# -*- coding: utf-8 -*-
'''
PowerGAMA
=========

Created on Fri Nov 01 13:09:04 2013
by Harald G Svendsen

Licence: `The MIT License (MIT) <http://opensource.org/licenses/MIT>`_

Classes and modules
-------------------
GridData (class)
    power grid model (nodes, branches, consumers, generators) and power inflow
    and storage value 
LpProblem (class)
    LP problem, with PuLP interface to external solver (COIN-CBC)
Results (class)
    Ouptut of optimisation, methods for analysis and plotting
constants
    global constants
scenarios
    methods to manipulate grid scenarios (generation and consumption)

Example
=======

>>> import powergama

Specify time range, i.e. which rows in the time series files to use, and 
number of hours per timestep

>>> timerange = range(0,8760)
>>> timedelta = 1.0

Import grid model and time series data from files

>>> data = powergama.GridData()
>>> data.readGridData(nodes="nodes.csv",
>>>					  ac_branches="branches.csv",
>>>                   dc_branches="hvdc.csv",
>>>                   generators="generators.csv",
>>>                   consumers="consumers.csv")
>>> data.readProfileData(inflow="inflow.csv",
>>>                      demand="demand.csv",
>>>                      storagevalue_filling="profile_storval_filling.csv",
>>>                      storagevalue_time="profile_storval_time.csv",
>>>                      timerange=timerange,
>>>                      timedelta=timedelta)                      

Initiate results and LP problem objects, including name of database file
(OBS: Will replace any existing file with the same name)

>>> lp = powergama.LpProblem(data)
>>> res = powergama.Results(data,"results.sqlite3")

Solve the problem, timestep by timestep, and save results in the results 
object

>>> lp.solve(res)

Plot results

>>> res.plotMapGrid(nodetype="nodalprice")

'''

from .GridData import GridData
from .LpProblemPyomo import LpProblem
from .Results import Results, ResultsBaseClass
